"""Purge cached files from all edge nodes."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('unique_id')
@click.argument('path')
@environment.pass_env
def cli(env, unique_id, path):
    """Creates a purge record and also initiates the purge call.

        Example:
             slcli cdn purge 9779455 /article/file.txt

        For more information see the following documentation: \n
        https://cloud.ibm.com/docs/infrastructure/CDN?topic=CDN-manage-your-cdn#purging-cached-content
    """

    manager = SoftLayer.CDNManager(env.client)
    result = manager.purge_content(unique_id, path)

    table = formatting.Table(['Date', 'Path', 'Saved', 'Status'])

    for data in result:
        table.add_row([
            data['date'],
            data['path'],
            data['saved'],
            data['status']
        ])

    env.fout(table)



"""Remove an origin pull mapping."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('unique_id')
@click.argument('origin_path')
@environment.pass_env
def cli(env, unique_id, origin_path):
    """Removes an origin path for an existing CDN mapping."""

    manager = SoftLayer.CDNManager(env.client)
    manager.remove_origin(unique_id, origin_path)

    click.secho("Origin with path %s has been deleted" % origin_path, fg='green')



"""Get details for an image."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer.CLI import image as image_mod
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Get details for an image."""

    image_mgr = SoftLayer.ImageManager(env.client)
    image_id = helpers.resolve_id(image_mgr.resolve_ids, identifier, 'image')
    image = image_mgr.get_image(image_id, mask=image_mod.DETAIL_MASK)

    children_images = image.get('children')
    total_size = utils.lookup(image, 'firstChild', 'blockDevicesDiskSpaceTotal') or 0

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    table.add_row(['id', image['id']])
    table.add_row(['global_identifier',
                   image.get('globalIdentifier', formatting.blank())])
    table.add_row(['name', image['name'].strip()])
    table.add_row(['status', formatting.FormattedItem(
        utils.lookup(image, 'status', 'keyname'),
        utils.lookup(image, 'status', 'name'),
    )])

    table.add_row([
        'active_transaction',
        formatting.listing(_get_transaction_groups(children_images), separator=','),
    ])
    table.add_row(['account', image.get('accountId', formatting.blank())])
    table.add_row(['visibility',
                   image_mod.PUBLIC_TYPE if image['publicFlag']
                   else image_mod.PRIVATE_TYPE])
    table.add_row(['type',
                   formatting.FormattedItem(
                       utils.lookup(image, 'imageType', 'keyName'),
                       utils.lookup(image, 'imageType', 'name'),
                   )])
    table.add_row(['flex', image.get('flexImageFlag')])
    table.add_row(['note', image.get('note')])
    table.add_row(['created', image.get('createDate')])
    table.add_row(['total_size', formatting.b_to_gb(total_size)])
    table.add_row(['datacenters', _get_datacenter_table(children_images)])

    env.fout(table)


def _get_datacenter_table(children_images):
    """Returns image details as datacenter, size, and transaction within a formatting table.

      :param children_images: A list of images.
      """
    table_datacenter = formatting.Table(['DC', 'size', 'transaction'])
    for child in children_images:
        table_datacenter.add_row([
            utils.lookup(child, 'datacenter', 'name'),
            formatting.b_to_gb(child.get('blockDevicesDiskSpaceTotal', 0)),
            formatting.transaction_status(child.get('transaction'))
        ])

    return table_datacenter


def _get_transaction_groups(children_images):
    """Returns a Set of transaction groups.

      :param children_images: A list of images.
      """
    transactions = set()
    for child in children_images:
        transactions.add(utils.lookup(child, 'transaction', 'transactionGroup', 'name'))

    return transactions



"""Request network configuration of an IPSEC tunnel context."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI.exceptions import CLIHalt


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('context_id', type=int)
@environment.pass_env
def cli(env, context_id):
    """Request configuration of a tunnel context.

    This action will update the advancedConfigurationFlag on the context
    instance and further modifications against the context will be prevented
    until all changes can be propgated to network devices.
    """
    manager = SoftLayer.IPSECManager(env.client)
    # ensure context can be retrieved by given id
    manager.get_tunnel_context(context_id)

    succeeded = manager.apply_configuration(context_id)
    if succeeded:
        click.echo('Configuration request received for context #{}'.format(context_id))
    else:
        raise CLIHalt('Failed to enqueue configuration request for context #{}'.format(context_id))



"""
    SoftLayer.cdn
    ~~~~~~~~~~~~~
    CDN Manager/helpers

    :license: MIT, see LICENSE for more details.
"""
import SoftLayer
from SoftLayer import utils


# pylint: disable=too-many-lines,too-many-instance-attributes


class CDNManager(utils.IdentifierMixin, object):
    """Manage Content Delivery Networks in the account.

    See product information here:
    https://www.ibm.com/cloud/cdn
    https://cloud.ibm.com/docs/infrastructure/CDN?topic=CDN-about-content-delivery-networks-cdn-

    :param SoftLayer.API.BaseClient client: the client instance
    """

    def __init__(self, client):
        self.client = client
        self._start_date = None
        self._end_date = None
        self.cdn_configuration = self.client['Network_CdnMarketplace_Configuration_Mapping']
        self.cdn_path = self.client['SoftLayer_Network_CdnMarketplace_Configuration_Mapping_Path']
        self.cdn_metrics = self.client['Network_CdnMarketplace_Metrics']
        self.cdn_purge = self.client['SoftLayer_Network_CdnMarketplace_Configuration_Cache_Purge']
        self.resolvers = [self._get_ids_from_hostname]

    def list_cdn(self, **kwargs):
        """Lists Content Delivery Networks for the active user.

        :param dict \\*\\*kwargs: header-level options (mask, limit, etc.)
        :returns: The list of CDN objects in the account
        """

        return self.cdn_configuration.listDomainMappings(**kwargs)

    def get_cdn(self, unique_id, **kwargs):
        """Retrieves the information about the CDN account object.

        :param str unique_id: The unique ID associated with the CDN.
        :param dict \\*\\*kwargs: header-level option (mask)
        :returns: The CDN object
        """

        cdn_list = self.cdn_configuration.listDomainMappingByUniqueId(unique_id, **kwargs)

        # The method listDomainMappingByUniqueId() returns an array but there is only 1 object
        return cdn_list[0]

    def get_origins(self, unique_id, **kwargs):
        """Retrieves list of origin pull mappings for a specified CDN account.

        :param str unique_id: The unique ID associated with the CDN.
        :param dict \\*\\*kwargs: header-level options (mask, limit, etc.)
        :returns: The list of origin paths in the CDN object.
        """

        return self.cdn_path.listOriginPath(unique_id, **kwargs)

    def add_origin(self, unique_id, origin, path, origin_type="server", header=None,
                   port=80, protocol='http', bucket_name=None, file_extensions=None,
                   optimize_for="web", cache_query="include all"):
        """Creates an origin path for an existing CDN.

        :param str unique_id: The unique ID associated with the CDN.
        :param str path: relative path to the domain provided, e.g. "/articles/video"
        :param str origin: ip address or hostname if origin_type=server, API endpoint for
                           your S3 object storage if origin_type=storage
        :param str origin_type: it can be 'server' or 'storage' types.
        :param str header: the edge server uses the host header to communicate with the origin.
                           It defaults to hostname. (optional)
        :param int port: the http port number (default: 80)
        :param str protocol: the protocol of the origin (default: HTTP)
        :param str bucket_name: name of the available resource
        :param str file_extensions: file extensions that can be stored in the CDN, e.g. "jpg,png"
        :param str optimize_for: performance configuration, available options: web, video, and file where:

                                    - 'web' = 'General web delivery'
                                    - 'video' = 'Video on demand optimization'
                                    - 'file' = 'Large file optimization'
        :param str cache_query: rules with the following formats: 'include-all', 'ignore-all',
                               'include: space separated query-names',
                               'ignore: space separated query-names'.'
        :return: a CDN origin path object
        """
        types = {'server': 'HOST_SERVER', 'storage': 'OBJECT_STORAGE'}
        performance_config = {
            'web': 'General web delivery',
            'video': 'Video on demand optimization',
            'file': 'Large file optimization'
        }

        new_origin = {
            'uniqueId': unique_id,
            'path': path,
            'origin': origin,
            'originType': types.get(origin_type),
            'httpPort': port,
            'protocol': protocol.upper(),
            'performanceConfiguration': performance_config.get(optimize_for, 'General web delivery'),
            'cacheKeyQueryRule': cache_query
        }

        if header:
            new_origin['header'] = header

        if types.get(origin_type) == 'OBJECT_STORAGE':
            if bucket_name:
                new_origin['bucketName'] = bucket_name

            if file_extensions:
                new_origin['fileExtension'] = file_extensions

        origin = self.cdn_path.createOriginPath(new_origin)

        # The method createOriginPath() returns an array but there is only 1 object
        return origin[0]

    def remove_origin(self, unique_id, path):
        """Removes an origin pull mapping with the given origin pull ID.

        :param str unique_id: The unique ID associated with the CDN.
        :param str path: The origin path to delete.
        :returns: A string value
        """

        return self.cdn_path.deleteOriginPath(unique_id, path)

    def purge_content(self, unique_id, path):
        """Purges a URL or path from the CDN.

        :param str unique_id: The unique ID associated with the CDN.
        :param str path: A string of url or path that should be purged.
        :returns: A Container_Network_CdnMarketplace_Configuration_Cache_Purge array object
        """

        return self.cdn_purge.createPurge(unique_id, path)

    def get_usage_metrics(self, unique_id, history=30, frequency="aggregate"):
        """Retrieves the cdn usage metrics.

        It uses the 'days' argument if start_date and end_date are None.

        :param int unique_id: The CDN uniqueId from which the usage metrics will be obtained.
        :param int history: Last N days, default days is 30.
        :param str frequency: It can be day, week, month and aggregate. The default is "aggregate".
        :returns: A Container_Network_CdnMarketplace_Metrics object
        """

        _start = utils.days_to_datetime(history)
        _end = utils.days_to_datetime(0)

        self._start_date = utils.timestamp(_start)
        self._end_date = utils.timestamp(_end)

        usage = self.cdn_metrics.getMappingUsageMetrics(unique_id, self._start_date, self._end_date, frequency)

        # The method getMappingUsageMetrics() returns an array but there is only 1 object
        return usage[0]

    @property
    def start_data(self):
        """Retrieve the cdn usage metric start date."""
        return self._start_date

    @property
    def end_date(self):
        """Retrieve the cdn usage metric end date."""
        return self._end_date

    def edit(self, identifier, header=None, http_port=None, https_port=None, origin=None,
             respect_headers=None, cache=None, performance_configuration=None):
        """Edit the cdn object.

        :param string identifier: The CDN identifier.
        :param header: The cdn Host header.
        :param http_port: The cdn HTTP port.
        :param https_port: The cdn HTTPS port.
        :param origin: The cdn Origin server address.
        :param respect_headers: The cdn Respect headers.
        :param cache: The cdn Cache key optimization.
        :param performance_configuration: The cdn performance configuration.

        :returns: SoftLayer_Container_Network_CdnMarketplace_Configuration_Mapping[].
        """
        cdn_instance_detail = self.get_cdn(str(identifier))

        config = {
            'uniqueId': cdn_instance_detail.get('uniqueId'),
            'originType': cdn_instance_detail.get('originType'),
            'protocol': cdn_instance_detail.get('protocol'),
            'path': cdn_instance_detail.get('path'),
            'vendorName': cdn_instance_detail.get('vendorName'),
            'cname': cdn_instance_detail.get('cname'),
            'domain': cdn_instance_detail.get('domain'),
            'origin': cdn_instance_detail.get('originHost'),
            'header': cdn_instance_detail.get('header')
        }
        if cdn_instance_detail.get('httpPort'):
            config['httpPort'] = cdn_instance_detail.get('httpPort')

        if cdn_instance_detail.get('httpsPort'):
            config['httpsPort'] = cdn_instance_detail.get('httpsPort')

        if header:
            config['header'] = header

        if http_port:
            config['httpPort'] = http_port

        if https_port:
            config['httpsPort'] = https_port

        if origin:
            config['origin'] = origin

        if respect_headers:
            config['respectHeaders'] = respect_headers

        if cache:
            if 'include-specified' in cache['cacheKeyQueryRule']:
                cache_key_rule = self.get_cache_key_query_rule('include', cache)
                config['cacheKeyQueryRule'] = cache_key_rule
            elif 'ignore-specified' in cache['cacheKeyQueryRule']:
                cache_key_rule = self.get_cache_key_query_rule('ignore', cache)
                config['cacheKeyQueryRule'] = cache_key_rule
            else:
                config['cacheKeyQueryRule'] = cache['cacheKeyQueryRule']

        if performance_configuration:
            config['performanceConfiguration'] = performance_configuration

        return self.cdn_configuration.updateDomainMapping(config)

    def _get_ids_from_hostname(self, hostname):
        """Get the cdn object detail.

        :param string hostname: The CDN identifier.
        :returns: SoftLayer_Container_Network_CdnMarketplace_Configuration_Mapping[].
        """
        result = []
        cdn_list = self.cdn_configuration.listDomainMappings()
        for cdn in cdn_list:
            if cdn.get('domain', '').lower() == hostname.lower():
                result.append(cdn.get('uniqueId'))
                break

        return result

    @staticmethod
    def get_cache_key_query_rule(cache_type, cache):
        """Get the cdn object detail.

        :param string cache_type: Cache type.
        :param  cache: Cache description.

        :return: string value.
        """
        if 'description' not in cache:
            raise SoftLayer.SoftLayerError('Please add a description to be able to update the'
                                           ' cache.')
        cache_result = '%s: %s' % (cache_type, cache['description'])

        return cache_result



"""List IPSEC VPN Tunnel Context Details."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('context_id', type=int)
@click.option('-i',
              '--include',
              default=[],
              multiple=True,
              type=click.Choice(['at', 'is', 'rs', 'sr', 'ss']),
              help='Include additional resources')
@environment.pass_env
def cli(env, context_id, include):
    """List IPSEC VPN tunnel context details.

    Additional resources can be joined using multiple instances of the
    include option, for which the following choices are available.

    \b
    at: address translations
    is: internal subnets
    rs: remote subnets
    sr: statically routed subnets
    ss: service subnets
    """
    mask = _get_tunnel_context_mask(('at' in include),
                                    ('is' in include),
                                    ('rs' in include),
                                    ('sr' in include),
                                    ('ss' in include))
    manager = SoftLayer.IPSECManager(env.client)
    context = manager.get_tunnel_context(context_id, mask=mask)

    env.fout(_get_context_table(context))

    for relation in include:
        if relation == 'at':
            env.fout(_get_address_translations_table(context.get('addressTranslations', [])))
        elif relation == 'is':
            env.fout(_get_subnets_table(context.get('internalSubnets', []), title="Internal Subnets"))
        elif relation == 'rs':
            env.fout(_get_subnets_table(context.get('customerSubnets', []), title="Remote Subnets"))
        elif relation == 'sr':
            env.fout(_get_subnets_table(context.get('staticRouteSubnets', []), title="Static Subnets"))
        elif relation == 'ss':
            env.fout(_get_subnets_table(context.get('serviceSubnets', []), title="Service Subnets"))


def _get_address_translations_table(address_translations):
    """Yields a formatted table to print address translations.

    :param List[dict] address_translations: List of address translations.
    :return Table: Formatted for address translation output.
    """
    table = formatting.Table(['id',
                              'static IP address',
                              'static IP address id',
                              'remote IP address',
                              'remote IP address id',
                              'note'], title="Address Translations")
    for address_translation in address_translations:
        table.add_row([address_translation.get('id', ''),
                       address_translation.get('internalIpAddressRecord', {}).get('ipAddress', ''),
                       address_translation.get('internalIpAddressId', ''),
                       address_translation.get('customerIpAddressRecord', {}).get('ipAddress', ''),
                       address_translation.get('customerIpAddressId', ''),
                       address_translation.get('notes', '')])
    return table


def _get_subnets_table(subnets, title):
    """Yields a formatted table to print subnet details.

    :param List[dict] subnets: List of subnets.
    :return Table: Formatted for subnet output.
    """
    table = formatting.Table(['id', 'network identifier', 'cidr', 'note'], title=title)
    for subnet in subnets:
        table.add_row([subnet.get('id', ''),
                       subnet.get('networkIdentifier', ''),
                       subnet.get('cidr', ''),
                       subnet.get('note', '')])
    return table


def _get_tunnel_context_mask(address_translations=False,
                             internal_subnets=False,
                             remote_subnets=False,
                             static_subnets=False,
                             service_subnets=False):
    """Yields a mask object for a tunnel context.

    All exposed properties on the tunnel context service are included in
    the constructed mask. Additional joins may be requested.

    :param bool address_translations: Whether to join the context's address
           translation entries.
    :param bool internal_subnets: Whether to join the context's internal
           subnet associations.
    :param bool remote_subnets: Whether to join the context's remote subnet
           associations.
    :param bool static_subnets: Whether to join the context's statically
           routed subnet associations.
    :param bool service_subnets: Whether to join the SoftLayer service
           network subnets.
    :return string: Encoding for the requested mask object.
    """
    entries = ['id',
               'accountId',
               'advancedConfigurationFlag',
               'createDate',
               'customerPeerIpAddress',
               'modifyDate',
               'name',
               'friendlyName',
               'internalPeerIpAddress',
               'phaseOneAuthentication',
               'phaseOneDiffieHellmanGroup',
               'phaseOneEncryption',
               'phaseOneKeylife',
               'phaseTwoAuthentication',
               'phaseTwoDiffieHellmanGroup',
               'phaseTwoEncryption',
               'phaseTwoKeylife',
               'phaseTwoPerfectForwardSecrecy',
               'presharedKey']
    if address_translations:
        entries.append('addressTranslations[internalIpAddressRecord[ipAddress],'
                       'customerIpAddressRecord[ipAddress]]')
    if internal_subnets:
        entries.append('internalSubnets')
    if remote_subnets:
        entries.append('customerSubnets')
    if static_subnets:
        entries.append('staticRouteSubnets')
    if service_subnets:
        entries.append('serviceSubnets')
    return '[mask[{}]]'.format(','.join(entries))


def _get_context_table(context):
    """Yields a formatted table to print context details.

    :param dict context: The tunnel context
    :return Table: Formatted for tunnel context output
    """
    table = formatting.KeyValueTable(['name', 'value'], title='Context Details')
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    table.add_row(['id', context.get('id', '')])
    table.add_row(['name', context.get('name', '')])
    table.add_row(['friendly name', context.get('friendlyName', '')])
    table.add_row(['internal peer IP address', context.get('internalPeerIpAddress', '')])
    table.add_row(['remote peer IP address', context.get('customerPeerIpAddress', '')])
    table.add_row(['advanced configuration flag', context.get('advancedConfigurationFlag', '')])
    table.add_row(['preshared key', context.get('presharedKey', '')])
    table.add_row(['phase 1 authentication', context.get('phaseOneAuthentication', '')])
    table.add_row(['phase 1 diffie hellman group', context.get('phaseOneDiffieHellmanGroup', '')])
    table.add_row(['phase 1 encryption', context.get('phaseOneEncryption', '')])
    table.add_row(['phase 1 key life', context.get('phaseOneKeylife', '')])
    table.add_row(['phase 2 authentication', context.get('phaseTwoAuthentication', '')])
    table.add_row(['phase 2 diffie hellman group', context.get('phaseTwoDiffieHellmanGroup', '')])
    table.add_row(['phase 2 encryption', context.get('phaseTwoEncryption', '')])
    table.add_row(['phase 2 key life', context.get('phaseTwoKeylife', '')])
    table.add_row(['phase 2 perfect forward secrecy', context.get('phaseTwoPerfectForwardSecrecy', '')])
    table.add_row(['created', context.get('createDate')])
    table.add_row(['modified', context.get('modifyDate')])
    return table



"""Cancel an IPSec service."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--immediate',
              is_flag=True,
              default=False,
              help="Cancels the service  immediately (instead of on the billing anniversary)")
@click.option('--reason',
              help="An optional cancellation reason. See cancel-reasons for a list of available options")
@environment.pass_env
def cli(env, identifier, immediate, reason):
    """Cancel a IPSEC VPN tunnel context."""

    manager = SoftLayer.IPSECManager(env.client)
    context = manager.get_tunnel_context(identifier, mask='billingItem')

    if 'billingItem' not in context:
        raise SoftLayer.SoftLayerError("Cannot locate billing. May already be cancelled.")

    result = manager.cancel_item(context['billingItem']['id'], immediate, reason)

    if result:
        env.fout("Ipsec {} was cancelled.".format(identifier))



"""Assigns the global IP to a target."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment

target_types = {'vlan': 'SoftLayer_Network_Vlan',
                'ip': 'SoftLayer_Network_Subnet_IpAddress',
                'hardware': 'SoftLayer_Hardware_Server',
                'vsi': 'SoftLayer_Virtual_Guest'}


@click.command(cls=SoftLayer.CLI.command.SLCommand, epilog="More information about types and identifiers "
               "on https://sldn.softlayer.com/reference/services/SoftLayer_Network_Subnet/route/")
@click.argument('identifier')
@click.option('--target', type=click.Choice(['vlan', 'ip', 'hardware', 'vsi']),
              help='choose the type. vlan, ip, hardware, vsi')
@click.option('--target-id', help='The identifier for the destination resource to route this subnet to. ')
@environment.pass_env
def cli(env, identifier, target, target_id):
    """Assigns the subnet to a target."""

    mgr = SoftLayer.NetworkManager(env.client)
    mgr.route(identifier, target_types.get(target), target_id)



listDomainMappings = [
    {
        "cacheKeyQueryRule": "include-all",
        "cname": "cdnakauuiet7s6u6.cdnedge.bluemix.net",
        "createDate": "2020-09-29T15:19:01-06:00",
        "domain": "test.example.com",
        "header": "test.example.com",
        "httpPort": 80,
        "originHost": "1.1.1.1",
        "originType": "HOST_SERVER",
        "path": "/",
        "protocol": "HTTP",
        "status": "CNAME_CONFIGURATION",
        "uniqueId": "11223344",
        "vendorName": "akamai"
    }
]

listDomainMappingByUniqueId = [
    {
        "cname": "cdnakauuiet7s6u6.cdnedge.bluemix.net",
        "performanceConfiguration": "Large file optimization",
        "domain": "test.example.com",
        "header": "test.example.com",
        "httpPort": 80,
        "originHost": "1.1.1.1",
        "originType": "HOST_SERVER",
        "path": "/",
        "protocol": "HTTP",
        "status": "CNAME_CONFIGURATION",
        "uniqueId": "11223344",
        "vendorName": "akamai"
    }
]

updateDomainMapping = [
    {
        "createDate": "2021-02-09T19:32:29-06:00",
        "originType": "HOST_SERVER",
        "path": "/*",
        "performanceConfiguration": "Large file optimization",
        "protocol": "HTTP",
        "respectHeaders": True,
        "uniqueId": "11223344",
        "vendorName": "akamai",
        "header": "www.test.com",
        "httpPort": 83,
        "cname": "cdn.test.cloud",
        "originHost": "1.1.1.1",
        "cacheKeyQueryRule": "include: test"
    }
]



"""Order a block storage volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions

CONTEXT_SETTINGS = {'token_normalize_func': lambda x: x.upper()}


@click.command(cls=SoftLayer.CLI.command.SLCommand, context_settings=CONTEXT_SETTINGS)
@click.option('--storage-type',
              help='Type of block storage volume',
              type=click.Choice(['performance', 'endurance']),
              required=True)
@click.option('--size',
              type=int,
              help='Size of block storage volume in GB.',
              required=True)
@click.option('--iops',
              type=int,
              help="""Performance Storage IOPs. Options vary based on storage size.
[required for storage-type performance]""")
@click.option('--tier',
              help='Endurance Storage Tier (IOP per GB) [required for storage-type endurance]',
              type=click.Choice(['0.25', '2', '4', '10']))
@click.option('--os-type',
              help='Operating System',
              type=click.Choice([
                  'HYPER_V',
                  'LINUX',
                  'VMWARE',
                  'WINDOWS_2008',
                  'WINDOWS_GPT',
                  'WINDOWS',
                  'XEN']),
              required=True)
@click.option('--location',
              help='Datacenter short name (e.g.: dal09)',
              required=True)
@click.option('--snapshot-size',
              type=int,
              help='Optional parameter for ordering snapshot '
              'space along with endurance block storage; specifies '
              'the size (in GB) of snapshot space to order')
@click.option('--service-offering',
              help="""The service offering package to use for placing the order.
[optional, default is \'storage_as_a_service\']. enterprise and performance are depreciated""",
              default='storage_as_a_service',
              type=click.Choice([
                  'storage_as_a_service',
                  'enterprise',
                  'performance']))
@click.option('--billing',
              type=click.Choice(['hourly', 'monthly']),
              default='monthly',
              help="Optional parameter for Billing rate (default to monthly)")
@environment.pass_env
def cli(env, storage_type, size, iops, tier, os_type,
        location, snapshot_size, service_offering, billing):
    """Order a block storage volume.

    Valid size and iops options can be found here:
    https://cloud.ibm.com/docs/BlockStorage/index.html#provisioning-considerations
    """
    block_manager = SoftLayer.BlockStorageManager(env.client)
    storage_type = storage_type.lower()

    hourly_billing_flag = False
    if billing.lower() == "hourly":
        hourly_billing_flag = True

    if service_offering != 'storage_as_a_service':
        click.secho('{} is a legacy storage offering'.format(service_offering), fg='red')
        if hourly_billing_flag:
            raise exceptions.CLIAbort(
                'Hourly billing is only available for the storage_as_a_service service offering'
            )

    if storage_type == 'performance':
        if iops is None:
            raise exceptions.CLIAbort('Option --iops required with Performance')

        if service_offering == 'performance' and snapshot_size is not None:
            raise exceptions.CLIAbort(
                '--snapshot-size is not available for performance service offerings. '
                'Use --service-offering storage_as_a_service'
            )

        try:
            order = block_manager.order_block_volume(
                storage_type=storage_type,
                location=location,
                size=int(size),
                iops=iops,
                os_type=os_type,
                snapshot_size=snapshot_size,
                service_offering=service_offering,
                hourly_billing_flag=hourly_billing_flag
            )
        except ValueError as ex:
            raise exceptions.ArgumentError(str(ex))

    if storage_type == 'endurance':
        if tier is None:
            raise exceptions.CLIAbort(
                'Option --tier required with Endurance in IOPS/GB [0.25,2,4,10]'
            )

        try:
            order = block_manager.order_block_volume(
                storage_type=storage_type,
                location=location,
                size=int(size),
                tier_level=float(tier),
                os_type=os_type,
                snapshot_size=snapshot_size,
                service_offering=service_offering,
                hourly_billing_flag=hourly_billing_flag
            )
        except ValueError as ex:
            raise exceptions.ArgumentError(str(ex))

    if 'placedOrder' in order.keys():
        click.echo("Order #{0} placed successfully!".format(
            order['placedOrder']['id']))
        for item in order['placedOrder']['items']:
            click.echo(" > %s" % item['description'])
        click.echo(
            '\nYou may run "slcli block volume-list --order {0}" to find this block volume after it '
            'is ready.'.format(order['placedOrder']['id']))
    else:
        click.echo("Order could not be placed! Please verify your options and try again.")



"""List virtual servers."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import columns as column_helper
from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer import utils

# pylint: disable=unnecessary-lambda

COLUMNS = [
    column_helper.Column('guid', ('globalIdentifier',)),
    column_helper.Column('primary_ip', ('primaryIpAddress',)),
    column_helper.Column('backend_ip', ('primaryBackendIpAddress',)),
    column_helper.Column('datacenter', ('datacenter', 'name')),
    column_helper.Column('action', lambda guest: formatting.active_txn(guest),
                         mask='''
                         activeTransaction[
                            id,transactionStatus[name,friendlyName]
                         ]'''),
    column_helper.Column('power_state', ('powerState', 'name')),
    column_helper.Column(
        'created_by',
        ('billingItem', 'orderItem', 'order', 'userRecord', 'username')),
    column_helper.Column(
        'tags',
        lambda server: formatting.tags(server.get('tagReferences')),
        mask="tagReferences.tag.name"),
]

DEFAULT_COLUMNS = [
    'id',
    'hostname',
    'domain',
    'deviceStatus.name',
    'datacenter',
    'primary_ip',
    'backend_ip',
    'createDate',
    'action',
]


@click.command(cls=SLCommand, short_help="List virtual servers.")
@click.option('--cpu', '-c', help='Number of CPU cores', type=click.INT)
@click.option('--domain', '-D', help='Domain portion of the FQDN')
@click.option('--datacenter', '-d', help='Datacenter shortname')
@click.option('--hostname', '-H', help='Host portion of the FQDN')
@click.option('--memory', '-m', help='Memory in mebibytes', type=click.INT)
@click.option('--network', '-n', help='Network port speed in Mbps')
@click.option('--hourly', is_flag=True, help='Show only hourly instances')
@click.option('--monthly', is_flag=True, help='Show only monthly instances')
@click.option('--transient', help='Filter by transient instances', type=click.BOOL)
@click.option('--hardware', is_flag=True, default=False, help='Show the all VSI related to hardware')
@click.option('--all-guests', is_flag=True, default=False, help='Show the all VSI and hardware VSIs')
@helpers.multi_option('--tag', help='Filter by tags')
@click.option('--sortby',
              help='Column to sort by',
              default='hostname',
              show_default=True)
@click.option('--columns',
              callback=column_helper.get_formatter(COLUMNS),
              help='Columns to display. [options: %s]'
              % ', '.join(column.name for column in COLUMNS),
              default=','.join(DEFAULT_COLUMNS),
              show_default=True)
@click.option('--limit', '-l',
              help='How many results to get in one api call, default is 100',
              default=100,
              show_default=True)
@environment.pass_env
def cli(env, sortby, cpu, domain, datacenter, hostname, memory, network,
        hourly, monthly, tag, columns, limit, transient, hardware, all_guests):
    """List virtual servers."""

    vsi = SoftLayer.VSManager(env.client)
    guests = vsi.list_instances(hourly=hourly,
                                monthly=monthly,
                                hostname=hostname,
                                domain=domain,
                                cpus=cpu,
                                memory=memory,
                                datacenter=datacenter,
                                nic_speed=network,
                                transient=transient,
                                tags=tag,
                                mask=columns.mask(),
                                limit=limit)

    table = formatting.Table(columns.columns)
    table.sortby = sortby
    if not hardware or all_guests:
        for guest in guests:
            table.add_row([value or formatting.blank()
                           for value in columns.row(guest)])

        env.fout(table)

    if hardware or all_guests:
        hardware_guests = vsi.get_hardware_guests()
        for hd_guest in hardware_guests:
            if hd_guest['virtualHost']['guests']:
                title = "Hardware(id = {hardwareId}) guests associated".format(hardwareId=hd_guest['id'])
                table_hardware_guest = formatting.Table(['id', 'hostname', 'CPU', 'Memory', 'Start Date', 'Status',
                                                         'powerState'], title=title)
                table_hardware_guest.sortby = 'hostname'
                for guest in hd_guest['virtualHost']['guests']:
                    table_hardware_guest.add_row([
                        guest['id'],
                        guest['hostname'],
                        '%i %s' % (guest['maxCpu'], guest['maxCpuUnits']),
                        guest['maxMemory'],
                        utils.clean_time(guest['createDate']),
                        guest['status']['keyName'],
                        guest['powerState']['keyName']
                    ])
                env.fout(table_hardware_guest)



"""Updates an IPSEC tunnel context."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI.exceptions import CLIHalt


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('context_id', type=int)
@click.option('--friendly-name',
              default=None,
              help='Friendly name value')
@click.option('--remote-peer',
              default=None,
              help='Remote peer IP address value')
@click.option('--preshared-key',
              default=None,
              help='Preshared key value')
@click.option('--phase1-auth',
              '--p1-auth',
              default=None,
              type=click.Choice(['MD5', 'SHA1', 'SHA256']),
              help='Phase 1 authentication value')
@click.option('--phase1-crypto',
              '--p1-crypto',
              default=None,
              type=click.Choice(['DES', '3DES', 'AES128', 'AES192', 'AES256']),
              help='Phase 1 encryption value')
@click.option('--phase1-dh',
              '--p1-dh',
              default=None,
              type=click.Choice(['0', '1', '2', '5']),
              help='Phase 1 diffie hellman group value')
@click.option('--phase1-key-ttl',
              '--p1-key-ttl',
              default=None,
              type=click.IntRange(120, 172800),
              help='Phase 1 key life value')
@click.option('--phase2-auth',
              '--p2-auth',
              default=None,
              type=click.Choice(['MD5', 'SHA1', 'SHA256']),
              help='Phase 2 authentication value')
@click.option('--phase2-crypto',
              '--p2-crypto',
              default=None,
              type=click.Choice(['DES', '3DES', 'AES128', 'AES192', 'AES256']),
              help='Phase 2 encryption value')
@click.option('--phase2-dh',
              '--p2-dh',
              default=None,
              type=click.Choice(['0', '1', '2', '5']),
              help='Phase 2 diffie hellman group value')
@click.option('--phase2-forward-secrecy',
              '--p2-forward-secrecy',
              default=None,
              type=click.IntRange(0, 1),
              help='Phase 2 perfect forward secrecy value')
@click.option('--phase2-key-ttl',
              '--p2-key-ttl',
              default=None,
              type=click.IntRange(120, 172800),
              help='Phase 2 key life value')
@environment.pass_env
def cli(env, context_id, friendly_name, remote_peer, preshared_key,
        phase1_auth, phase1_crypto, phase1_dh, phase1_key_ttl, phase2_auth,
        phase2_crypto, phase2_dh, phase2_forward_secrecy, phase2_key_ttl):
    """Update tunnel context properties.

    Updates are made atomically, so either all are accepted or none are.

    Key life values must be in the range 120-172800.

    Phase 2 perfect forward secrecy must be in the range 0-1.

    A separate configuration request should be made to realize changes on
    network devices.
    """
    manager = SoftLayer.IPSECManager(env.client)
    succeeded = manager.update_tunnel_context(
        context_id,
        friendly_name=friendly_name,
        remote_peer=remote_peer,
        preshared_key=preshared_key,
        phase1_auth=phase1_auth,
        phase1_crypto=phase1_crypto,
        phase1_dh=phase1_dh,
        phase1_key_ttl=phase1_key_ttl,
        phase2_auth=phase2_auth,
        phase2_crypto=phase2_crypto,
        phase2_dh=phase2_dh,
        phase2_forward_secrecy=phase2_forward_secrecy,
        phase2_key_ttl=phase2_key_ttl
    )
    if succeeded:
        env.out('Updated context #{}'.format(context_id))
    else:
        raise CLIHalt('Failed to update context #{}'.format(context_id))



"""Delete autoscale."""
# :license: MIT, see LICENSE for more details.
import click
import SoftLayer

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.managers.autoscale import AutoScaleManager


@click.command(cls=SLCommand)
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Delete this group and destroy all members of it.

    Example: slcli autoscale delete autoscaleId

    """

    autoscale = AutoScaleManager(env.client)

    try:
        autoscale.delete(identifier)
        click.secho("%s deleted successfully" % identifier, fg='green')
    except SoftLayer.SoftLayerAPIError as ex:
        click.secho("Failed to delete %s\n%s" % (identifier, ex), fg='red')



"""
    SoftLayer.ipsec
    ~~~~~~~~~~~~~~~~~~
    IPSec VPN Manager

    :license: MIT, see LICENSE for more details.
"""

from SoftLayer.exceptions import SoftLayerAPIError
from SoftLayer.managers import ordering
from SoftLayer import utils


class IPSECManager(utils.IdentifierMixin, object):
    """Manage SoftLayer IPSEC VPN tunnel contexts.

    This provides helpers to manage IPSEC contexts, private and remote subnets,
    and NAT translations.

    :param SoftLayer.API.BaseClient client: the client instance
    :param SoftLayer.API.BaseClient account: account service client
    :param SoftLayer.API.BaseClient context: tunnel context client
    :param SoftLayer.API.BaseClient customer_subnet: remote subnet client
    """

    def __init__(self, client):
        self.client = client
        self.account = client['Account']
        self.context = client['Network_Tunnel_Module_Context']
        self.remote_subnet = client['Network_Customer_Subnet']

    def add_internal_subnet(self, context_id, subnet_id):
        """Add an internal subnet to a tunnel context.

        :param int context_id: The id-value representing the context instance.
        :param int subnet_id: The id-value representing the internal subnet.
        :return bool: True if internal subnet addition was successful.
        """
        return self.context.addPrivateSubnetToNetworkTunnel(subnet_id,
                                                            id=context_id)

    def add_remote_subnet(self, context_id, subnet_id):
        """Adds a remote subnet to a tunnel context.

        :param int context_id: The id-value representing the context instance.
        :param int subnet_id: The id-value representing the remote subnet.
        :return bool: True if remote subnet addition was successful.
        """
        return self.context.addCustomerSubnetToNetworkTunnel(subnet_id,
                                                             id=context_id)

    def add_service_subnet(self, context_id, subnet_id):
        """Adds a service subnet to a tunnel context.

        :param int context_id: The id-value representing the context instance.
        :param int subnet_id: The id-value representing the service subnet.
        :return bool: True if service subnet addition was successful.
        """
        return self.context.addServiceSubnetToNetworkTunnel(subnet_id,
                                                            id=context_id)

    def apply_configuration(self, context_id):
        """Requests network configuration for a tunnel context.

        :param int context_id: The id-value representing the context instance.
        :return bool: True if the configuration request was successfully queued.
        """
        return self.context.applyConfigurationsToDevice(id=context_id)

    def create_remote_subnet(self, account_id, identifier, cidr):
        """Creates a remote subnet on the given account.

        :param string account_id: The account identifier.
        :param string identifier: The network identifier of the remote subnet.
        :param string cidr: The CIDR value of the remote subnet.
        :return dict: Mapping of properties for the new remote subnet.
        """
        return self.remote_subnet.createObject({
            'accountId': account_id,
            'cidr': cidr,
            'networkIdentifier': identifier
        })

    def create_translation(self, context_id, static_ip, remote_ip, notes):
        """Creates an address translation on a tunnel context/

        :param int context_id: The id-value representing the context instance.
        :param string static_ip: The IP address value representing the
               internal side of the translation entry,
        :param string remote_ip: The IP address value representing the remote
               side of the translation entry,
        :param string notes: The notes to supply with the translation entry,
        :return dict: Mapping of properties for the new translation entry.
        """
        return self.context.createAddressTranslation({
            'customerIpAddress': remote_ip,
            'internalIpAddress': static_ip,
            'notes': notes
        }, id=context_id)

    def delete_remote_subnet(self, subnet_id):
        """Deletes a remote subnet from the current account.

        :param string subnet_id: The id-value representing the remote subnet.
        :return bool: True if subnet deletion was successful.
        """
        return self.remote_subnet.deleteObject(id=subnet_id)

    def get_tunnel_context(self, context_id, **kwargs):
        """Retrieves the network tunnel context instance.

        :param int context_id: The id-value representing the context instance.
        :return dict: Mapping of properties for the tunnel context.
        :raise SoftLayerAPIError: If a context cannot be found.
        """
        _filter = utils.NestedDict(kwargs.get('filter') or {})
        _filter['networkTunnelContexts']['id'] = utils.query_filter(context_id)

        kwargs['filter'] = _filter.to_dict()
        contexts = self.account.getNetworkTunnelContexts(**kwargs)
        if len(contexts) == 0:
            raise SoftLayerAPIError('SoftLayer_Exception_ObjectNotFound',
                                    'Unable to find object with id of \'{}\''
                                    .format(context_id))
        return contexts[0]

    def get_translation(self, context_id, translation_id):
        """Retrieves a translation entry for the given id values.

        :param int context_id: The id-value representing the context instance.
        :param int translation_id: The id-value representing the translation
               instance.
        :return dict: Mapping of properties for the translation entry.
        :raise SoftLayerAPIError: If a translation cannot be found.
        """
        translation = next((x for x in self.get_translations(context_id)
                            if x['id'] == translation_id), None)
        if translation is None:
            raise SoftLayerAPIError('SoftLayer_Exception_ObjectNotFound',
                                    'Unable to find object with id of \'{}\''
                                    .format(translation_id))
        return translation

    def get_translations(self, context_id):
        """Retrieves all translation entries for a tunnel context.

        :param int context_id: The id-value representing the context instance.
        :return list(dict): Translations associated with the given context
        """
        _mask = ('[mask[addressTranslations[customerIpAddressRecord,'
                 'internalIpAddressRecord]]]')
        context = self.get_tunnel_context(context_id, mask=_mask)
        # Pull the internal and remote IP addresses into the translation
        for translation in context.get('addressTranslations', []):
            remote_ip = translation.get('customerIpAddressRecord', {})
            internal_ip = translation.get('internalIpAddressRecord', {})
            translation['customerIpAddress'] = remote_ip.get('ipAddress', '')
            translation['internalIpAddress'] = internal_ip.get('ipAddress', '')
            translation.pop('customerIpAddressRecord', None)
            translation.pop('internalIpAddressRecord', None)
        return context['addressTranslations']

    def get_tunnel_contexts(self, **kwargs):
        """Retrieves network tunnel module context instances.

        :return list(dict): Contexts associated with the current account.
        """
        return self.account.getNetworkTunnelContexts(**kwargs)

    def remove_internal_subnet(self, context_id, subnet_id):
        """Remove an internal subnet from a tunnel context.

        :param int context_id: The id-value representing the context instance.
        :param int subnet_id: The id-value representing the internal subnet.
        :return bool: True if internal subnet removal was successful.
        """
        return self.context.removePrivateSubnetFromNetworkTunnel(subnet_id,
                                                                 id=context_id)

    def remove_remote_subnet(self, context_id, subnet_id):
        """Removes a remote subnet from a tunnel context.

        :param int context_id: The id-value representing the context instance.
        :param int subnet_id: The id-value representing the remote subnet.
        :return bool: True if remote subnet removal was successful.
        """
        return self.context.removeCustomerSubnetFromNetworkTunnel(subnet_id,
                                                                  id=context_id)

    def remove_service_subnet(self, context_id, subnet_id):
        """Removes a service subnet from a tunnel context.

        :param int context_id: The id-value representing the context instance.
        :param int subnet_id: The id-value representing the service subnet.
        :return bool: True if service subnet removal was successful.
        """
        return self.context.removeServiceSubnetFromNetworkTunnel(subnet_id,
                                                                 id=context_id)

    def remove_translation(self, context_id, translation_id):
        """Removes a translation entry from a tunnel context.

        :param int context_id: The id-value representing the context instance.
        :param int translation_id: The id-value representing the translation.
        :return bool: True if translation entry removal was successful.
        """
        return self.context.deleteAddressTranslation(translation_id,
                                                     id=context_id)

    def update_translation(self, context_id, translation_id, static_ip=None,
                           remote_ip=None, notes=None):
        """Updates an address translation entry using the given values.

        :param int context_id: The id-value representing the context instance.
        :param dict template: A key-value mapping of translation properties.
        :param string static_ip: The static IP address value to update.
        :param string remote_ip: The remote IP address value to update.
        :param string notes: The notes value to update.
        :return bool: True if the update was successful.
        """
        translation = self.get_translation(context_id, translation_id)

        if static_ip is not None:
            translation['internalIpAddress'] = static_ip
            translation.pop('internalIpAddressId', None)
        if remote_ip is not None:
            translation['customerIpAddress'] = remote_ip
            translation.pop('customerIpAddressId', None)
        if notes is not None:
            translation['notes'] = notes
        self.context.editAddressTranslation(translation, id=context_id)
        return True

    def update_tunnel_context(self, context_id, friendly_name=None,
                              remote_peer=None, preshared_key=None,
                              phase1_auth=None, phase1_crypto=None,
                              phase1_dh=None, phase1_key_ttl=None,
                              phase2_auth=None, phase2_crypto=None,
                              phase2_dh=None, phase2_forward_secrecy=None,
                              phase2_key_ttl=None):
        """Updates a tunnel context using the given values.

        :param string context_id: The id-value representing the context.
        :param string friendly_name: The friendly name value to update.
        :param string remote_peer: The remote peer IP address value to update.
        :param string preshared_key: The preshared key value to update.
        :param string phase1_auth: The phase 1 authentication value to update.
        :param string phase1_crypto: The phase 1 encryption value to update.
        :param string phase1_dh: The phase 1 diffie hellman group value
               to update.
        :param string phase1_key_ttl: The phase 1 key life value to update.
        :param string phase2_auth: The phase 2 authentication value to update.
        :param string phase2_crypto: The phase 2 encryption value to update.
        :param string phase2_df: The phase 2 diffie hellman group value
               to update.
        :param string phase2_forward_secriecy: The phase 2 perfect forward
               secrecy value to update.
        :param string phase2_key_ttl: The phase 2 key life value to update.
        :return bool: True if the update was successful.
        """
        context = self.get_tunnel_context(context_id)

        if friendly_name is not None:
            context['friendlyName'] = friendly_name
        if remote_peer is not None:
            context['customerPeerIpAddress'] = remote_peer
        if preshared_key is not None:
            context['presharedKey'] = preshared_key
        if phase1_auth is not None:
            context['phaseOneAuthentication'] = phase1_auth
        if phase1_crypto is not None:
            context['phaseOneEncryption'] = phase1_crypto
        if phase1_dh is not None:
            context['phaseOneDiffieHellmanGroup'] = phase1_dh
        if phase1_key_ttl is not None:
            context['phaseOneKeylife'] = phase1_key_ttl
        if phase2_auth is not None:
            context['phaseTwoAuthentication'] = phase2_auth
        if phase2_crypto is not None:
            context['phaseTwoEncryption'] = phase2_crypto
        if phase2_dh is not None:
            context['phaseTwoDiffieHellmanGroup'] = phase2_dh
        if phase2_forward_secrecy is not None:
            context['phaseTwoPerfectForwardSecrecy'] = phase2_forward_secrecy
        if phase2_key_ttl is not None:
            context['phaseTwoKeylife'] = phase2_key_ttl
        return self.context.editObject(context, id=context_id)

    def order(self, datacenter, item_package):
        """Create a ipsec.

        :param string datacenter: the datacenter shortname
        :param string[] item_package: items array
        """
        complex_type = 'SoftLayer_Container_Product_Order_Network_Tunnel_Ipsec'
        ordering_manager = ordering.OrderingManager(self.client)
        return ordering_manager.place_order(package_keyname='ADDITIONAL_PRODUCTS',
                                            location=datacenter,
                                            item_keynames=item_package,
                                            complex_type=complex_type,
                                            hourly=False)

    def cancel_item(self, identifier, immediate, reason):
        """Cancels the specified billing item Ipsec.

        Example::

            # Cancels ipsec id 1234
            result = mgr.cancel_item(billing_item_id=1234)

        :param int billing_id: The ID of the billing item to be cancelled.
        :param string reason: The reason code for the cancellation. This should come from
                              :func:`get_cancellation_reasons`.
        :param bool immediate: If set to True, will automatically update the cancelation ticket to request
                               the resource be reclaimed asap. This request still has to be reviewed by a human
        :returns: True on success or an exception
        """
        return self.client.call('SoftLayer_Billing_Item', 'cancelItem',
                                True, immediate, reason, id=identifier)



"""Detail a CDN Account."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('unique_id')
@click.option('--history',
              default=30, type=click.IntRange(1, 89),
              help='Bandwidth, Hits, Ratio counted over history number of days ago. 89 is the maximum. ')
@environment.pass_env
def cli(env, unique_id, history):
    """Detail a CDN Account."""

    manager = SoftLayer.CDNManager(env.client)

    cdn_mapping = manager.get_cdn(unique_id)
    cdn_metrics = manager.get_usage_metrics(unique_id, history=history)

    # usage metrics
    total_bandwidth = "%s GB" % cdn_metrics['totals'][0]
    total_hits = cdn_metrics['totals'][1]
    hit_ratio = "%s %%" % cdn_metrics['totals'][2]

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    table.add_row(['unique_id', cdn_mapping['uniqueId']])
    table.add_row(['hostname', cdn_mapping['domain']])
    table.add_row(['protocol', cdn_mapping['protocol']])
    table.add_row(['origin', cdn_mapping['originHost']])
    table.add_row(['origin_type', cdn_mapping['originType']])
    table.add_row(['path', cdn_mapping['path']])
    table.add_row(['provider', cdn_mapping['vendorName']])
    table.add_row(['status', cdn_mapping['status']])
    table.add_row(['total_bandwidth', total_bandwidth])
    table.add_row(['total_hits', total_hits])
    table.add_row(['hit_ratio', hit_ratio])

    env.fout(table)



"""Remove a translation entry from an IPSEC tunnel context."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI.exceptions import CLIHalt


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('context_id', type=int)
@click.option('-t',
              '--translation-id',
              required=True,
              type=int,
              help='Translation identifier to remove')
@environment.pass_env
def cli(env, context_id, translation_id):
    """Remove a translation entry from an IPSEC tunnel context.

    A separate configuration request should be made to realize changes on
    network devices.
    """
    manager = SoftLayer.IPSECManager(env.client)
    # ensure translation can be retrieved by given id
    manager.get_translation(context_id, translation_id)

    succeeded = manager.remove_translation(context_id, translation_id)
    if succeeded:
        env.out('Removed translation #{}'.format(translation_id))
    else:
        raise CLIHalt('Failed to remove translation #{}'.format(translation_id))



"""Metric Utilities"""
import datetime
import itertools
import sys

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils


# pylint: disable=unused-argument
def _validate_datetime(ctx, param, value):
    try:
        return datetime.datetime.strptime(value, "%Y-%m-%d")
    except (ValueError, TypeError):
        pass

    try:
        return datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
    except (ValueError, TypeError) as ex:
        message = "not in the format 'YYYY-MM-DD' or 'YYYY-MM-DD HH:MM:SS'"
        raise click.BadParameter(message) from ex


def _get_pooled_bandwidth(env, start, end):
    call = env.client.call('Account', 'getVirtualDedicatedRacks',
                           iter=True,
                           mask='id,name,metricTrackingObjectId')
    types = [
        {'keyName': 'PUBLICIN',
         'name': 'publicIn',
         'summaryType': 'sum'},
        {'keyName': 'PUBLICOUT',
         'name': 'publicOut',
         'summaryType': 'sum'},
        {'keyName': 'PRIVATEIN',
         'name': 'privateIn',
         'summaryType': 'sum'},
        {'keyName': 'PRIVATEOUT',
         'name': 'privateOut',
         'summaryType': 'sum'},
    ]

    with click.progressbar(list(call),
                           label='Calculating for bandwidth pools',
                           file=sys.stderr) as pools:
        for pool in pools:
            metric_tracking_id = pool.get('metricTrackingObjectId')

            if metric_tracking_id is None:
                continue

            pool_detail = {
                'id': pool.get('id'),
                'type': 'pool',
                'name': pool.get('name'),
                'data': []
            }
            bw_data = env.client.call(
                'Metric_Tracking_Object',
                'getSummaryData',
                start.strftime('%Y-%m-%d %H:%M:%S %Z'),
                end.strftime('%Y-%m-%d %H:%M:%S %Z'),
                types,
                300,
                id=metric_tracking_id,
            )
            pool_detail['data'] = bw_data

            yield pool_detail


def _get_hardware_bandwidth(env, start, end):
    hw_call = env.client.call(
        'Account', 'getHardware',
        iter=True,
        mask='id,hostname,metricTrackingObject.id,'
             'virtualRack[id,bandwidthAllotmentTypeId,name]')
    types = [
        {'keyName': 'PUBLICIN',
         'name': 'publicIn',
         'summaryType': 'counter'},
        {'keyName': 'PUBLICOUT',
         'name': 'publicOut',
         'summaryType': 'counter'},
        {'keyName': 'PRIVATEIN',
         'name': 'privateIn',
         'summaryType': 'counter'},
        {'keyName': 'PRIVATEOUT',
         'name': 'privateOut',
         'summaryType': 'counter'},
    ]

    with click.progressbar(list(hw_call),
                           label='Calculating for hardware',
                           file=sys.stderr) as hws:
        for instance in hws:
            if not utils.lookup(instance, 'metricTrackingObject', 'id'):
                continue

            pool_name = None
            if utils.lookup(instance,
                            'virtualRack',
                            'bandwidthAllotmentTypeId') == 2:
                pool_name = utils.lookup(instance, 'virtualRack', 'name')

            yield {
                'id': instance['id'],
                'type': 'hardware',
                'name': instance['hostname'],
                'pool': pool_name,
                'data': env.client.call(
                    'Metric_Tracking_Object',
                    'getSummaryData',
                    start.strftime('%Y-%m-%d %H:%M:%S %Z'),
                    end.strftime('%Y-%m-%d %H:%M:%S %Z'),
                    types,
                    3600,
                    id=instance['metricTrackingObject']['id'],
                ),
            }


def _get_virtual_bandwidth(env, start, end):
    call = env.client.call(
        'Account', 'getVirtualGuests',
        iter=True,
        mask='id,hostname,metricTrackingObjectId,'
             'virtualRack[id,bandwidthAllotmentTypeId,name]')
    types = [
        {'keyName': 'PUBLICIN_NET_OCTET',
         'name': 'publicIn_net_octet',
         'summaryType': 'sum'},
        {'keyName': 'PUBLICOUT_NET_OCTET',
         'name': 'publicOut_net_octet',
         'summaryType': 'sum'},
        {'keyName': 'PRIVATEIN_NET_OCTET',
         'name': 'privateIn_net_octet',
         'summaryType': 'sum'},
        {'keyName': 'PRIVATEOUT_NET_OCTET',
         'name': 'privateOut_net_octet',
         'summaryType': 'sum'},
    ]

    with click.progressbar(list(call),
                           label='Calculating for virtual',
                           file=sys.stderr) as vms:
        for instance in vms:
            metric_tracking_id = utils.lookup(instance,
                                              'metricTrackingObjectId')

            if metric_tracking_id is None:
                continue

            pool_name = None
            if utils.lookup(instance,
                            'virtualRack',
                            'bandwidthAllotmentTypeId') == 2:
                pool_name = utils.lookup(instance, 'virtualRack', 'name')

            yield {
                'id': instance['id'],
                'type': 'virtual',
                'name': instance['hostname'],
                'pool': pool_name,
                'data': env.client.call(
                    'Metric_Tracking_Object',
                    'getSummaryData',
                    start.strftime('%Y-%m-%d %H:%M:%S %Z'),
                    end.strftime('%Y-%m-%d %H:%M:%S %Z'),
                    types,
                    3600,
                    id=metric_tracking_id,
                ),
            }


@click.command(cls=SLCommand, short_help="Bandwidth report for every pool/server")
@click.option('--start', callback=_validate_datetime,
              default=(datetime.datetime.now() - datetime.timedelta(days=30)).strftime('%Y-%m-%d'),
              help="datetime in the format 'YYYY-MM-DD' or 'YYYY-MM-DD HH:MM:SS'")
@click.option('--end', callback=_validate_datetime, default=datetime.datetime.now().strftime('%Y-%m-%d'),
              help="datetime in the format 'YYYY-MM-DD' or 'YYYY-MM-DD HH:MM:SS'")
@click.option('--sortby', help='Column to sort by', default='hostname', show_default=True)
@click.option('--virtual', is_flag=True, default=False,
              help='Show only the bandwidth summary for each virtual server')
@click.option('--server', is_flag=True, default=False,
              help='Show only the bandwidth summary for each hardware server')
@click.option('--pool', is_flag=True, default=False,
              help='Show only the bandwidth pool summary.')
@environment.pass_env
def cli(env, start, end, sortby, virtual, server, pool):
    """Bandwidth report for every pool/server.

    This reports on the total data transfered for each virtual sever, hardware
    server and bandwidth pool.
    """

    env.err('Generating bandwidth report for %s to %s' % (start, end))

    table = formatting.Table([
        'type',
        'hostname',
        'public_in',
        'public_out',
        'private_in',
        'private_out',
        'pool',
    ])
    table.sortby = sortby

    def f_type(key, results):
        "Filter metric data by type"
        return (result['counter'] for result in results
                if result['type'] == key)

    def _input_to_table(item):
        "Input metric data to table"
        pub_in = int(sum(f_type('publicIn_net_octet', item['data'])))
        pub_out = int(sum(f_type('publicOut_net_octet', item['data'])))
        pri_in = int(sum(f_type('privateIn_net_octet', item['data'])))
        pri_out = int(sum(f_type('privateOut_net_octet', item['data'])))
        table.add_row([
            item['type'],
            item['name'],
            formatting.b_to_gb(pub_in),
            formatting.b_to_gb(pub_out),
            formatting.b_to_gb(pri_in),
            formatting.b_to_gb(pri_out),
            item.get('pool') or formatting.blank(),
        ])

    try:
        if virtual:
            for item in itertools.chain(_get_pooled_bandwidth(env, start, end),
                                        _get_virtual_bandwidth(env, start, end)):
                _input_to_table(item)
        elif server:
            for item in itertools.chain(_get_pooled_bandwidth(env, start, end),
                                        _get_hardware_bandwidth(env, start, end)):
                _input_to_table(item)
        elif pool:
            for item in _get_pooled_bandwidth(env, start, end):
                _input_to_table(item)
        else:
            for item in itertools.chain(_get_pooled_bandwidth(env, start, end),
                                        _get_hardware_bandwidth(env, start, end),
                                        _get_virtual_bandwidth(env, start, end)):
                _input_to_table(item)
    except KeyboardInterrupt:
        env.err("Printing collected results and then aborting.")

    env.out(env.fmt(table))



STAAS_TEST_VOLUME = {
    'accountId': 1234,
    'activeTransactions': None,
    'activeTransactionCount': 0,
    'billingItem': {
        'activeChildren': [{
            'categoryCode': 'storage_snapshot_space',
            'id': 125,
            'cancellationDate': '',
        }],
        'cancellationDate': '',
        'categoryCode': 'storage_as_a_service',
        'hourlyFlag': None,
        'id': 454,
        'location': {'id': 449500}
    },
    'capacityGb': 500,
    'hasEncryptionAtRest': 1,
    'id': 102,
    'iops': 1000,
    'lunId': 2,
    'osType': {'keyName': 'LINUX'},
    'originalVolumeSize': '500',
    'parentVolume': {'snapshotSizeBytes': 1024},
    'provisionedIops': '1000',
    'replicationPartnerCount': 0,
    'schedules': [{
        'id': 978,
        'type': {'keyname': 'SNAPSHOT_WEEKLY'},
    }],
    'serviceResource': {'datacenter': {'id': 449500, 'name': 'dal05'}},
    'serviceResourceBackendIpAddress': '10.1.2.3',
    'snapshotCapacityGb': '10',
    'staasVersion': '2',
    'storageTierLevel': 'READHEAVY_TIER',
    'storageType': {'keyName': 'ENDURANCE_BLOCK_STORAGE'},
    'username': 'duplicatable_volume_username'
}

getObject = {
    'accountId': 1234,
    'activeTransactionCount': 1,
    'activeTransactions': [{
        'transactionStatus': {'friendlyName': 'This is a buffer time in which the customer may cancel the server'}
    }],
    'allowedHardware': [{
        'allowedHost': {
            'credential': {'username': 'joe', 'password': '12345'},
            'name': 'test-server',
        },
        'domain': 'example.com',
        'hostname': 'test-server',
        'id': 1234,
        'primaryBackendIpAddress': '10.0.0.2',
    }],
    'allowedIpAddresses': [{
        'allowedHost': {
            'credential': {'username': 'joe', 'password': '12345'},
            'name': 'test-server',
        },
        'id': 1234,
        'ipAddress': '10.0.0.1',
        'note': 'backend ip',
    }],
    'allowedSubnets': [{
        'allowedHost': {
            'credential': {'username': 'joe', 'password': '12345'},
            'name': 'test-server',
        },
        'cidr': '24',
        'id': 1234,
        'networkIdentifier': '10.0.0.1',
        'note': 'backend subnet',
    }],
    'allowedVirtualGuests': [{
        'allowedHost': {
            'credential': {'username': 'joe', 'password': '12345'},
            'name': 'test-server',
        },
        'domain': 'example.com',
        'hostname': 'test-server',
        'id': 1234,
        'primaryBackendIpAddress': '10.0.0.1',
    }],
    'billingItem': {
        'activeChildren': [{
            'cancellationDate': '',
            'categoryCode': 'storage_snapshot_space',
            'id': 123,
        }],
        'cancellationDate': '',
        'categoryCode': 'storage_service_enterprise',
        'id': 449,
        'location': {'id': 449500}
    },
    'bytesUsed': 0,
    'capacityGb': 20,
    'createDate': '2015:50:15-04:00',
    'fileNetworkMountAddress': '127.0.0.1:/TEST',
    'guestId': '',
    'hardwareId': '',
    'hasEncryptionAtRest': 0,
    'hostId': '',
    'id': 100,
    'iops': 1000,
    'lunId': 2,
    'nasType': 'ISCSI',
    'notes': """{'status': 'available'}""",
    'originalSnapshotName': 'test-original-snapshot-name',
    'originalVolumeName': 'test-original-volume-name',
    'originalVolumeSize': '20',
    'osType': {'keyName': 'LINUX'},
    'parentVolume': {'snapshotSizeBytes': 1024},
    'password': '',
    'provisionedIops': '1000',
    'replicationPartnerCount': 1,
    'replicationPartners': [{
        'createDate': '2017:50:15-04:00',
        'id': 1784,
        'nasType': 'ISCSI_REPLICANT',
        'replicationSchedule': {'type': {'keyname': 'REPLICATION_HOURLY'}},
        'serviceResource': {'datacenter': {'name': 'wdc01'}},
        'serviceResourceBackendIpAddress': '10.3.174.79',
        'username': 'TEST_REP_1',
    }, {
        'createDate': '2017:50:15-04:00',
        'id': 1785,
        'nasType': 'ISCSI_REPLICANT',
        'replicationSchedule': {'type': {'keyname': 'REPLICATION_DAILY'}},
        'serviceResource': {'datacenter': {'name': 'dal01'}},
        'serviceResourceBackendIpAddress': '10.3.177.84',
        'username': 'TEST_REP_2',
    }],
    'replicationStatus': 'Replicant Volume Provisioning has completed.',
    'schedules': [
        {
            'id': 978,
            'type': {'keyname': 'SNAPSHOT_WEEKLY'},
            'properties': [
                {'type': {'keyname': 'MINUTE'}, 'value': '30'},
            ]
        },
        {
            'id': 988,
            'type': {'keyname': 'REPLICATION_INTERVAL'},
            'properties': [
                {'type': {'keyname': 'MINUTE'}, 'value': '-1'},
            ]
        }
    ],
    'serviceProviderId': 1,
    'serviceResource': {'datacenter': {'id': 449500, 'name': 'dal05'},
                        'name': 'Cleversafe - US Region',
                        'type': {
                            'type': 'CLEVERSAFE_SVC_API'
                        }},
    'serviceResourceBackendIpAddress': '10.1.2.3',
    'serviceResourceName': 'Storage Type 01 Aggregate staaspar0101_pc01',
    'snapshotCapacityGb': '10',
    'staasVersion': '1',
    'storageTierLevel': 'READHEAVY_TIER',
    'storageType': {'keyName': 'ENDURANCE_STORAGE'},
    'username': 'username',
    'dependentDuplicate': 1,
}

getSnapshots = [{
    'id': 470,
    'notes': 'unit_testing_note',
    'snapshotCreationTimestamp': '2016-07-06T07:41:19-05:00',
    'snapshotSizeBytes': '42',
}]

getReplicationPartners = [{
    'id': 1784,
    'accountId': 3000,
    'capacityGb': 20,
    'username': 'TEST_REP_1',
    'serviceResourceBackendIpAddress': '10.3.174.79',
    'nasType': 'ISCSI_REPLICANT',
    'hostId': None,
    'guestId': None,
    'hardwareId': None,
    'createDate': '2017:50:15-04:00',
    'serviceResource': {'datacenter': {'name': 'wdc01'}},
    'replicationSchedule': {'type': {'keyname': 'REPLICATION_HOURLY'}},
}, {
    'id': 1785,
    'accountId': 3001,
    'capacityGb': 20,
    'username': 'TEST_REP_2',
    'serviceResourceBackendIpAddress': '10.3.177.84',
    'nasType': 'ISCSI_REPLICANT',
    'hostId': None,
    'guestId': None,
    'hardwareId': None,
    'createDate': '2017:50:15-04:00',
    'serviceResource': {'datacenter': {'name': 'dal01'}},
    'replicationSchedule': {'type': {'keyname': 'REPLICATION_DAILY'}},
}]

getValidReplicationTargetDatacenterLocations = [{
    'id': 12345,
    'longName': 'Dallas 05',
    'name': 'dal05'
}]

listVolumeSchedules = [
    {
        'id': 978,
        'type': {'keyname': 'SNAPSHOT_WEEKLY'},
        'properties': [{'type': {'keyname': 'MINUTE'}, 'value': '30'}]
    },
    {
        'id': 988,
        'type': {'keyname': 'REPLICATION_INTERVAL'},
        'properties': [{'type': {'keyname': 'MINUTE'}, 'value': '-1'}]
    }
]

deleteObject = True
editObject = True
allowAccessFromHostList = True
removeAccessFromHostList = True
failoverToReplicant = True
failbackFromReplicant = True
restoreFromSnapshot = True
disasterRecoveryFailoverToReplicant = True

createSnapshot = {
    'id': 449
}

enableSnapshots = True
disableSnapshots = True

getVolumeCountLimits = {
    'datacenterName': 'global',
    'maximumAvailableCount': 300,
    'provisionedCount': 100
}

refreshDuplicate = {
    'dependentDuplicate': 1
}

convertCloneDependentToIndependent = {
    'dependentDuplicate': 1
}



"""Add an address translation to an IPSEC tunnel context."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
# from SoftLayer.CLI.exceptions import ArgumentError
# from SoftLayer.CLI.exceptions import CLIHalt


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('context_id', type=int)
@click.option('-s',
              '--static-ip',
              required=True,
              help='Static IP address value')
@click.option('-r',
              '--remote-ip',
              required=True,
              help='Remote IP address value')
@click.option('-n',
              '--note',
              default=None,
              help='Note value')
@environment.pass_env
def cli(env, context_id, static_ip, remote_ip, note):
    """Add an address translation to an IPSEC tunnel context.

    A separate configuration request should be made to realize changes on
    network devices.
    """
    manager = SoftLayer.IPSECManager(env.client)
    # ensure context can be retrieved by given id
    manager.get_tunnel_context(context_id)

    translation = manager.create_translation(context_id,
                                             static_ip=static_ip,
                                             remote_ip=remote_ip,
                                             notes=note)
    env.out('Created translation from {} to {} #{}'
            .format(static_ip, remote_ip, translation['id']))



"""Order a IPSec VPN tunnel."""
# :licenses: MIT, see LICENSE for more details.

import click

import SoftLayer

from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--datacenter', '-d', required=True, prompt=True, help="Datacenter shortname")
@environment.pass_env
def cli(env, datacenter):
    """Order/create a IPSec VPN tunnel instance."""

    ipsec_manager = SoftLayer.IPSECManager(env.client)

    if not (env.skip_confirmations or formatting.confirm(
            "This action will incur charges on your account. Continue?")):
        raise exceptions.CLIAbort('Aborting ipsec order.')

    result = ipsec_manager.order(datacenter, ['IPSEC_STANDARD'])

    table = formatting.KeyValueTable(['Name', 'Value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table.add_row(['Id', result['orderId']])
    table.add_row(['Created', result['orderDate']])
    table.add_row(['Name', result['placedOrder']['items'][0]['description']])

    env.fout(table)



"""
    SoftLayer.license
    ~~~~~~~~~~~~~~~
    License Manager
    :license: MIT, see LICENSE for more details.
"""

# pylint: disable=too-many-public-methods

LICENSE_PACKAGE_ID = 301


class LicensesManager(object):
    """Manages account lincese."""

    def __init__(self, client):
        self.client = client

    def get_create_options(self):
        """Returns valid options for ordering Licenses.

        """

        return self.client.call('SoftLayer_Product_Package', 'getItems',
                                id=LICENSE_PACKAGE_ID)



"""List IPSec VPN Tunnel Contexts."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@environment.pass_env
def cli(env):
    """List IPSec VPN tunnel contexts"""
    manager = SoftLayer.IPSECManager(env.client)
    contexts = manager.get_tunnel_contexts()

    table = formatting.Table(['id',
                              'name',
                              'friendly name',
                              'internal peer IP address',
                              'remote peer IP address',
                              'created'])
    for context in contexts:
        table.add_row([context.get('id', ''),
                       context.get('name', ''),
                       context.get('friendlyName', ''),
                       context.get('internalPeerIpAddress', ''),
                       context.get('customerPeerIpAddress', ''),
                       context.get('createDate', '')])
    env.fout(table)



getObject = {
    'primaryRouter': {
        'datacenter': {'id': 1234, 'longName': 'TestDC'},
        'fullyQualifiedDomainName': 'fcr01.TestDC'
    },
    'id': 1234,
    'vlanNumber': 4444,
    'firewallInterfaces': None,
    'billingItem': {
        'allowCancellationFlag': 1,
        'categoryCode': 'network_vlan',
        'description': 'Private Network Vlan',
        'id': 235689,
        'notes': 'test cli',
        'orderItemId': 147258,
    },
    'subnets': [
        {
            'broadcastAddress': '169.46.22.127',
            'cidr': 28,
            'gateway': '169.46.22.113',
            'id': 1804813,
            'netmask': '255.255.255.240',
            'networkIdentifier': '169.46.22.112',
            'networkVlanId': 1404267,
            'subnetType': 'ADDITIONAL_PRIMARY',
            'totalIpAddresses': '16',
            'usableIpAddressCount': '13',
            'addressSpace': 'PUBLIC'
        },
        {
            'broadcastAddress': '150.239.7.191',
            'cidr': 27,
            'gateway': '150.239.7.161',
            'id': 2415982,
            'netmask': '255.255.255.224',
            'networkIdentifier': '150.239.7.160',
            'networkVlanId': 1404267,
            'subnetType': 'SECONDARY_ON_VLAN',
            'totalIpAddresses': '32',
            'usableIpAddressCount': '29',
            'version': 4,
            'addressSpace': 'PUBLIC'
        }],
    'hardware': [],
    'virtualGuests': [],
    'tagReferences': [],
}

editObject = True
setTags = True
getList = [getObject]

cancel = True

getCancelFailureReasons = [
    "1 bare metal server(s) still on the VLAN ",
    "1 virtual guest(s) still on the VLAN "
]



getInvoiceTopLevelItems = [
    {
        'categoryCode': 'sov_sec_ip_addresses_priv',
        'createDate': '2018-04-04T23:15:20-06:00',
        'description': '64 Portable Private IP Addresses',
        'id': 724951323,
        'oneTimeAfterTaxAmount': '0',
        'recurringAfterTaxAmount': '0',
        'hostName': 'bleg',
        'domainName': 'beh.com',
        'category': {'name': 'Private (only) Secondary VLAN IP Addresses'},
        'children': [
            {
                'id': 12345,
                'category': {'name': 'Fake Child Category'},
                'description': 'Blah',
                'oneTimeAfterTaxAmount': 55.50,
                'recurringAfterTaxAmount': 0.10
            }
        ],
        'location': {'name': 'fra02'}
    }
]



"""Invoice details"""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.account import AccountManager as AccountManager
from SoftLayer import utils


@click.command(cls=SLCommand)
@click.argument('identifier')
@click.option('--details', is_flag=True, default=False, show_default=True,
              help="Shows a very detailed list of charges")
@environment.pass_env
def cli(env, identifier, details):
    """Invoice details"""

    manager = AccountManager(env.client)
    top_items = manager.get_billing_items(identifier)
    table = get_invoice_table(identifier, top_items, details)
    env.fout(table)


def nice_string(ugly_string, limit=100):
    """Format and trims strings"""
    return (ugly_string[:limit] + '..') if len(ugly_string) > limit else ugly_string


def get_invoice_table(identifier, top_items, details):
    """Formats a table for invoice top level items.

     :param int identifier: Invoice identifier.
     :param list top_items: invoiceTopLevelItems.
     :param bool details: To add very detailed list of charges.
      """

    title = "Invoice %s" % identifier
    table = formatting.Table(["Item Id", "Category", "Description", "Single",
                              "Monthly", "Create Date", "Location"], title=title)
    table.align['category'] = 'l'
    table.align['description'] = 'l'
    for item in top_items:
        fqdn = "%s.%s" % (item.get('hostName', ''), item.get('domainName', ''))
        # category id=2046, ram_usage doesn't have a name...
        category = utils.lookup(item, 'category', 'name') or item.get('categoryCode')
        description = nice_string(item.get('description'))
        if fqdn != '.':
            description = "%s (%s)" % (item.get('description'), fqdn)
        table.add_row([
            item.get('id'),
            category,
            nice_string(description),
            "$%.2f" % float(item.get('oneTimeAfterTaxAmount')),
            "$%.2f" % float(item.get('recurringAfterTaxAmount')),
            utils.clean_time(item.get('createDate'), out_format="%Y-%m-%d"),
            utils.lookup(item, 'location', 'name')
        ])
        if details:
            for child in item.get('children', []):
                table.add_row([
                    '>>>',
                    utils.lookup(child, 'category', 'name'),
                    nice_string(child.get('description')),
                    "$%.2f" % float(child.get('oneTimeAfterTaxAmount')),
                    "$%.2f" % float(child.get('recurringAfterTaxAmount')),
                    '---',
                    '---'
                ])
    return table



"""Add a subnet to an IPSEC tunnel context."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI.custom_types import NetworkParamType
from SoftLayer.CLI import environment
from SoftLayer.CLI.exceptions import ArgumentError
from SoftLayer.CLI.exceptions import CLIHalt


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('context_id', type=int)
@click.option('-s',
              '--subnet-id',
              default=None,
              type=int,
              help='Subnet identifier to add')
@click.option('-t',
              '--subnet-type',
              '--type',
              required=True,
              type=click.Choice(['internal', 'remote', 'service']),
              help='Subnet type to add')
@click.option('-n',
              '--network-identifier',
              '--network',
              default=None,
              type=NetworkParamType(),
              help='Subnet network identifier to create')
@environment.pass_env
def cli(env, context_id, subnet_id, subnet_type, network_identifier):
    """Add a subnet to an IPSEC tunnel context.

    A subnet id may be specified to link to the existing tunnel context.

    Otherwise, a network identifier in CIDR notation should be specified,
    indicating that a subnet resource should first be created before associating
    it with the tunnel context. Note that this is only supported for remote
    subnets, which are also deleted upon failure to attach to a context.

    A separate configuration request should be made to realize changes on
    network devices.
    """
    create_remote = False
    if subnet_id is None:
        if network_identifier is None:
            raise ArgumentError('Either a network identifier or subnet id '
                                'must be provided.')
        if subnet_type != 'remote':
            raise ArgumentError('Unable to create {} subnets'
                                .format(subnet_type))
        create_remote = True

    manager = SoftLayer.IPSECManager(env.client)
    context = manager.get_tunnel_context(context_id)

    if create_remote:
        subnet = manager.create_remote_subnet(context['accountId'],
                                              identifier=network_identifier[0],
                                              cidr=network_identifier[1])
        subnet_id = subnet['id']
        env.out('Created subnet {}/{} #{}'
                .format(network_identifier[0],
                        network_identifier[1],
                        subnet_id))

    succeeded = False
    if subnet_type == 'internal':
        succeeded = manager.add_internal_subnet(context_id, subnet_id)
    elif subnet_type == 'remote':
        succeeded = manager.add_remote_subnet(context_id, subnet_id)
    elif subnet_type == 'service':
        succeeded = manager.add_service_subnet(context_id, subnet_id)

    if succeeded:
        env.out('Added {} subnet #{}'.format(subnet_type, subnet_id))
    else:
        raise CLIHalt('Failed to add {} subnet #{}'
                      .format(subnet_type, subnet_id))



"""
    SoftLayer.search
    ~~~~~~~~~~~~~~~~~~
    Search Manager

    :license: MIT, see LICENSE for more details.
"""


class SearchManager(object):
    """Manager to help searcha via the SoftLayer API.

    :param SoftLayer.API.BaseClient client: the client instance
    """

    def __init__(self, client):
        self.client = client
        self.search_manager = client['SoftLayer_Search']

    def get_object_types(self):
        """returns a collection of SoftLayer_Container_Search_ObjectType containers.

        """
        return self.search_manager.getObjectTypes()

    def search(self, search_string):
        """allows for searching for SoftLayer resources by simple phrase.

        """
        return self.search_manager.search(search_string)

    def advanced(self, search_string):
        """allows for searching for SoftLayer resources by simple phrase.

        """
        return self.search_manager.advancedSearch(search_string)



"""Remove a subnet from an IPSEC tunnel context."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI.exceptions import CLIHalt


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('context_id', type=int)
@click.option('-s',
              '--subnet-id',
              required=True,
              type=int,
              help='Subnet identifier to remove')
@click.option('-t',
              '--subnet-type',
              '--type',
              required=True,
              type=click.Choice(['internal', 'remote', 'service']),
              help='Subnet type to add')
@environment.pass_env
def cli(env, context_id, subnet_id, subnet_type):
    """Remove a subnet from an IPSEC tunnel context.

    The subnet id to remove must be specified.

    Remote subnets are deleted upon removal from a tunnel context.

    A separate configuration request should be made to realize changes on
    network devices.
    """
    manager = SoftLayer.IPSECManager(env.client)
    # ensure context can be retrieved by given id
    manager.get_tunnel_context(context_id)

    succeeded = False
    if subnet_type == 'internal':
        succeeded = manager.remove_internal_subnet(context_id, subnet_id)
    elif subnet_type == 'remote':
        succeeded = manager.remove_remote_subnet(context_id, subnet_id)
    elif subnet_type == 'service':
        succeeded = manager.remove_service_subnet(context_id, subnet_id)

    if succeeded:
        env.out('Removed {} subnet #{}'.format(subnet_type, subnet_id))
    else:
        raise CLIHalt('Failed to remove {} subnet #{}'
                      .format(subnet_type, subnet_id))



"""Edit a CDN Account."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--header', '-H',
              type=click.STRING,
              help="Host header."
              )
@click.option('--http-port', '-t',
              type=click.INT,
              help="HTTP port."
              )
@click.option('--https-port', '-s',
              type=click.INT,
              help="HTTPS port."
              )
@click.option('--origin', '-o',
              type=click.STRING,
              help="Origin server address."
              )
@click.option('--respect-headers', '-r',
              type=click.Choice(['1', '0']),
              help="Respect headers. The value 1 is On and 0 is Off."
              )
@click.option('--cache', '-c', multiple=True, type=str,
              help="Cache key optimization. These are the valid options to choose: 'include-all', 'ignore-all', "
                   "'include-specified', 'ignore-specified'. If you select 'include-specified' or 'ignore-specified' "
                   "please add a description too using again --cache, "
                   "e.g --cache=include-specified --cache=description."
              )
@click.option('--performance-configuration', '-p',
              type=click.Choice(['General web delivery', 'Large file optimization', 'Video on demand optimization']),
              help="Optimize for, General web delivery', 'Large file optimization', 'Video on demand optimization', "
                   "the Dynamic content acceleration option is not added because this has a special configuration."
              )
@environment.pass_env
def cli(env, identifier, header, http_port, https_port, origin, respect_headers, cache, performance_configuration):
    """Edit a CDN Account.

       Note: You can use the hostname or uniqueId as IDENTIFIER.
    """

    manager = SoftLayer.CDNManager(env.client)
    cdn_id = helpers.resolve_id(manager.resolve_ids, identifier, 'CDN')

    cache_result = {}
    if cache:
        if len(cache) > 1:
            cache_result['cacheKeyQueryRule'] = cache[0]
            cache_result['description'] = cache[1]
        else:
            cache_result['cacheKeyQueryRule'] = cache[0]

    cdn_result = manager.edit(cdn_id, header=header, http_port=http_port, https_port=https_port, origin=origin,
                              respect_headers=respect_headers, cache=cache_result,
                              performance_configuration=performance_configuration)

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    for cdn in cdn_result:
        table.add_row(['Create Date', cdn.get('createDate')])
        table.add_row(['Header', cdn.get('header')])
        if cdn.get('httpPort'):
            table.add_row(['Http Port', cdn.get('httpPort')])
        if cdn.get('httpsPort'):
            table.add_row(['Https Port', cdn.get('httpsPort')])
        table.add_row(['Origin Type', cdn.get('originType')])
        table.add_row(['Performance Configuration', cdn.get('performanceConfiguration')])
        table.add_row(['Protocol', cdn.get('protocol')])
        table.add_row(['Respect Headers', cdn.get('respectHeaders')])
        table.add_row(['Unique Id', cdn.get('uniqueId')])
        table.add_row(['Vendor Name', cdn.get('vendorName')])
        table.add_row(['Cache key optimization', cdn.get('cacheKeyQueryRule')])
        table.add_row(['cname', cdn.get('cname')])
        table.add_row(['Origin server address', cdn.get('originHost')])

    env.fout(table)



"""List number of file storage volumes limit per datacenter."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting

DEFAULT_COLUMNS = [
    'Datacenter',
    'MaximumAvailableCount',
    'ProvisionedCount'
]


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--sortby', help='Column to sort by', default='Datacenter')
@environment.pass_env
def cli(env, sortby):
    """List number of block storage volumes limit per datacenter."""
    file_manager = SoftLayer.FileStorageManager(env.client)
    file_volumes = file_manager.list_file_volume_limit()

    table = formatting.KeyValueTable(DEFAULT_COLUMNS)
    table.sortby = sortby
    for volume in file_volumes:
        datacenter_name = volume['datacenterName']
        maximum_available_count = volume['maximumAvailableCount']
        provisioned_count = volume['provisionedCount']
        table.add_row([datacenter_name, maximum_available_count, provisioned_count])
    env.fout(table)



"""List images."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import image as image_mod
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--name', default=None, help='Filter on image name')
@click.option('--public/--private', is_flag=True, default=None,
              help='Display only public or private images')
@environment.pass_env
def cli(env, name, public):
    """List images."""

    image_mgr = SoftLayer.ImageManager(env.client)

    images = []
    if public in [False, None]:
        for image in image_mgr.list_private_images(name=name, mask=image_mod.MASK):
            images.append(image)

    if public in [True, None]:
        for image in image_mgr.list_public_images(name=name, mask=image_mod.MASK):
            images.append(image)

    table = formatting.Table(['id', 'name', 'type', 'visibility', 'account'])

    images = [image for image in images if not image['parentId']]
    for image in images:

        visibility = (image_mod.PUBLIC_TYPE if image['publicFlag'] else image_mod.PRIVATE_TYPE)
        table.add_row([
            image.get('id', formatting.blank()),
            formatting.FormattedItem(image['name'], click.wrap_text(image['name'], width=50)),
            formatting.FormattedItem(
                utils.lookup(image, 'imageType', 'keyName'),
                utils.lookup(image, 'imageType', 'name')),
            visibility,
            image.get('accountId', formatting.blank()),
        ])

    env.fout(table)



"""List origin pull mappings."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('unique_id')
@environment.pass_env
def cli(env, unique_id):
    """List origin path for an existing CDN mapping."""

    manager = SoftLayer.CDNManager(env.client)
    origins = manager.get_origins(unique_id)

    table = formatting.Table(['Path', 'Origin', 'HTTP Port', 'Status'])

    for origin in origins:
        table.add_row([origin['path'],
                       origin['origin'],
                       origin['httpPort'],
                       origin['status']])

    env.fout(table)



"""Create a zone."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('zone')
@environment.pass_env
def cli(env, zone):
    """Create a zone."""

    manager = SoftLayer.DNSManager(env.client)
    manager.create_zone(zone)



"""Get details of an Autoscale groups."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.autoscale import AutoScaleManager
from SoftLayer import utils


@click.command(cls=SLCommand)
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Get details of an Autoscale group."""

    autoscale = AutoScaleManager(env.client)
    group = autoscale.details(identifier)

    # Group Config Table
    table = formatting.KeyValueTable(["Group", "Value"])
    table.align['Group'] = 'l'
    table.align['Value'] = 'l'

    table.add_row(['Id', group.get('id')])
    # Ideally we would use regionalGroup->preferredDatacenter, but that generates an error.
    table.add_row(['Datacenter', group['regionalGroup']['locations'][0]['longName']])
    table.add_row(['Termination', utils.lookup(group, 'terminationPolicy', 'name')])
    table.add_row(['Minimum Members', group.get('minimumMemberCount')])
    table.add_row(['Maximum Members', group.get('maximumMemberCount')])
    table.add_row(['Current Members', group.get('virtualGuestMemberCount')])
    table.add_row(['Cooldown', "{} seconds".format(group.get('cooldown'))])
    table.add_row(['Last Action', utils.clean_time(group.get('lastActionDate'))])

    for network in group.get('networkVlans', []):
        network_type = utils.lookup(network, 'networkVlan', 'networkSpace')
        router = utils.lookup(network, 'networkVlan', 'primaryRouter', 'hostname')
        vlan_number = utils.lookup(network, 'networkVlan', 'vlanNumber')
        vlan_name = "{}.{}".format(router, vlan_number)
        table.add_row([network_type, vlan_name])

    env.fout(table)

    # Template Config Table
    config_table = formatting.KeyValueTable(["Template", "Value"])
    config_table.align['Template'] = 'l'
    config_table.align['Value'] = 'l'

    template = group.get('virtualGuestMemberTemplate')

    config_table.add_row(['Hostname', template.get('hostname')])
    config_table.add_row(['Domain', template.get('domain')])
    config_table.add_row(['Core', template.get('startCpus')])
    config_table.add_row(['Ram', template.get('maxMemory')])
    network = template.get('networkComponents')
    config_table.add_row(['Network', network[0]['maxSpeed'] if network else 'Default'])
    ssh_keys = template.get('sshKeys', [])
    ssh_manager = SoftLayer.SshKeyManager(env.client)
    for key in ssh_keys:
        # Label isn't included when retrieved from the AutoScale group...
        ssh_key = ssh_manager.get_key(key.get('id'))
        config_table.add_row(['SSH Key {}'.format(ssh_key.get('id')), ssh_key.get('label')])
    disks = template.get('blockDevices', [])
    disk_type = "Local" if template.get('localDiskFlag') else "SAN"

    for disk in disks:
        disk_image = disk.get('diskImage')
        config_table.add_row(['{} Disk {}'.format(disk_type, disk.get('device')), disk_image.get('capacity')])
    config_table.add_row(['OS', template.get('operatingSystemReferenceCode')])
    config_table.add_row(['Post Install', template.get('postInstallScriptUri') or 'None'])

    env.fout(config_table)

    # Policy Config Table
    policy_table = formatting.KeyValueTable(["Policy", "Cooldown"])
    policies = group.get('policies', [])
    for policy in policies:
        policy_table.add_row([policy.get('name'), policy.get('cooldown') or group.get('cooldown')])

    env.fout(policy_table)

    # Active Guests
    member_table = formatting.Table(['Id', 'Hostname', 'Created'], title="Active Guests")
    guests = group.get('virtualGuestMembers', [])
    for guest in guests:
        real_guest = guest.get('virtualGuest')
        member_table.add_row([
            real_guest.get('id'), real_guest.get('hostname'), utils.clean_time(real_guest.get('provisionDate'))
        ])
    env.fout(member_table)



"""
    SoftLayer.autoscale
    ~~~~~~~~~~~~~~~~~~~
    Autoscale manager

    :license: MIT, see LICENSE for more details.
"""


class AutoScaleManager(object):
    """Manager for interacting with Autoscale instances."""

    def __init__(self, client):
        self.client = client

    def list(self, mask=None):
        """Calls `SoftLayer_Account::getScaleGroups()`_

        :param mask: optional SoftLayer_Scale_Group objectMask

        .. _SoftLayer_Account::getScaleGroups():
            https://sldn.softlayer.com/reference/services/SoftLayer_Account/getScaleGroups/
        """
        if not mask:
            mask = "mask[status,virtualGuestMemberCount]"

        return self.client.call('SoftLayer_Account', 'getScaleGroups', mask=mask, iter=True)

    def details(self, identifier, mask=None):
        """Calls `SoftLayer_Scale_Group::getObject()`_

        :param identifier: SoftLayer_Scale_Group id
        :param mask: optional SoftLayer_Scale_Group objectMask

        .. _SoftLayer_Scale_Group::getObject():
            https://sldn.softlayer.com/reference/services/SoftLayer_Scale_Group/getObject/
        """
        if not mask:
            mask = """mask[virtualGuestMembers[id,virtualGuest[id,hostname,domain,provisionDate]], terminationPolicy,
                   virtualGuestMemberCount, virtualGuestMemberTemplate[sshKeys],
                   policies[id,name,createDate,cooldown,actions,triggers,scaleActions],
                   networkVlans[networkVlanId,networkVlan[networkSpace,primaryRouter[hostname]]],
                   loadBalancers, regionalGroup[locations]]"""
        return self.client.call('SoftLayer_Scale_Group', 'getObject', id=identifier, mask=mask)

    def get_policy(self, identifier, mask=None):
        """Calls `SoftLayer_Scale_Policy::getObject()`_

        :param identifier: SoftLayer_Scale_Policy id
        :param mask: optional SoftLayer_Scale_Policy objectMask

        .. _SoftLayer_Scale_Policy::getObject():
            https://sldn.softlayer.com/reference/services/SoftLayer_Scale_Policy/getObject/
        """
        if not mask:
            mask = """mask[cooldown, createDate, id, name, actions, triggers[type]

            ]"""

        return self.client.call('SoftLayer_Scale_Policy', 'getObject', id=identifier, mask=mask)

    def scale(self, identifier, amount):
        """Calls `SoftLayer_Scale_Group::scale()`_

        :param identifier: SoftLayer_Scale_Group Id
        :param amount: positive or negative number to scale the group by

        .. _SoftLayer_Scale_Group::scale():
            https://sldn.softlayer.com/reference/services/SoftLayer_Scale_Group/scale/
        """
        return self.client.call('SoftLayer_Scale_Group', 'scale', amount, id=identifier)

    def scale_to(self, identifier, amount):
        """Calls `SoftLayer_Scale_Group::scaleTo()`_

        :param identifier: SoftLayer_Scale_Group Id
        :param amount: number to scale the group to.

        .. _SoftLayer_Scale_Group::scaleTo():
            https://sldn.softlayer.com/reference/services/SoftLayer_Scale_Group/scaleTo/
        """
        return self.client.call('SoftLayer_Scale_Group', 'scaleTo', amount, id=identifier)

    def get_logs(self, identifier, mask=None, object_filter=None):
        """Calls `SoftLayer_Scale_Group::getLogs()`_

        :param identifier: SoftLayer_Scale_Group Id
        :param mask: optional SoftLayer_Scale_Group_Log objectMask
        :param object_filter: optional SoftLayer_Scale_Group_Log objectFilter

        .. _SoftLayer_Scale_Group::getLogs():
            https://sldn.softlayer.com/reference/services/SoftLayer_Scale_Group/getLogs/
        """
        return self.client.call('SoftLayer_Scale_Group', 'getLogs', id=identifier, mask=mask, filter=object_filter,
                                iter=True)

    def get_virtual_guests(self, identifier, mask=None):
        """Calls `SoftLayer_Scale_Group::getVirtualGuestMembers()`_

        :param identifier: SoftLayer_Scale_Group Id
        :param mask: optional SoftLayer_Scale_Member objectMask

        .. _SoftLayer_Scale_Group::getVirtualGuestMembers():
            https://sldn.softlayer.com/reference/services/SoftLayer_Scale_Group/getVirtualGuestMembers/
        """
        return self.client.call('SoftLayer_Scale_Group', 'getVirtualGuestMembers', id=identifier, mask=mask, iter=True)

    def edit(self, identifier, template):
        """Calls `SoftLayer_Scale_Group::editObject()`_

        :param identifier: SoftLayer_Scale_Group id
        :param template: `SoftLayer_Scale_Group`_

        .. _SoftLayer_Scale_Group::editObject():
            https://sldn.softlayer.com/reference/services/SoftLayer_Scale_Group/editObject/
        .. _SoftLayer_Scale_Group: https://sldn.softlayer.com/reference/datatypes/SoftLayer_Scale_Group/
        """
        return self.client.call('SoftLayer_Scale_Group', 'editObject', template, id=identifier)

    def create(self, template):
        """Calls `SoftLayer_Scale_Group::createObject()`_

               :param template: `SoftLayer_Scale_Group`_

                .. _SoftLayer_Scale_Group::createObject():
                    https://sldn.softlayer.com/reference/services/SoftLayer_Scale_Group/createObject/
                .. _SoftLayer_Scale_Group: https://sldn.softlayer.com/reference/datatypes/SoftLayer_Scale_Group/
                """
        return self.client.call('SoftLayer_Scale_Group', 'createObject', template)

    def delete(self, identifier):
        """Calls `SoftLayer_Scale_Group::forceDeleteObject()`_

        :param identifier: SoftLayer_Scale_Group id

        .. _SoftLayer_Scale_Group::forceDeleteObject():
            https://sldn.softlayer.com/reference/services/SoftLayer_Scale_Group/forceDeleteObject/
        """
        return self.client.call('SoftLayer_Scale_Group', 'forceDeleteObject', id=identifier)



"""List Autoscale groups."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.autoscale import AutoScaleManager
from SoftLayer import utils


@click.command(cls=SLCommand)
@environment.pass_env
def cli(env):
    """List all Autoscale Groups on your account."""

    autoscale = AutoScaleManager(env.client)
    groups = autoscale.list()

    table = formatting.Table(["Id", "Name", "Status", "Min/Max", "Running"])
    table.align['Name'] = 'l'
    for group in groups:
        status = utils.lookup(group, 'status', 'name')
        min_max = "{}/{}".format(group.get('minimumMemberCount'), group.get('maximumMemberCount'))
        table.add_row([
            group.get('id'), group.get('name'), status, min_max, group.get('virtualGuestMemberCount')
        ])

    env.fout(table)



"""Create an origin pull mapping."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('unique_id')
@click.argument('origin')
@click.argument('path')
@click.option('--origin-type', '-t',
              type=click.Choice(['server', 'storage']),
              help='The origin type.',
              default='server',
              show_default=True)
@click.option('--header', '-H',
              type=click.STRING,
              help='The host header to communicate with the origin.')
@click.option('--bucket-name', '-b',
              type=click.STRING,
              help="The name of the available resource [required if --origin-type=storage]")
@click.option('--port', '-p',
              type=click.INT,
              help="The http port number.",
              default=80,
              show_default=True)
@click.option('--protocol', '-P',
              type=click.STRING,
              help="The protocol used by the origin.",
              default='http',
              show_default=True)
@click.option('--optimize-for', '-o',
              type=click.Choice(['web', 'video', 'file']),
              help="Performance configuration",
              default='web',
              show_default=True)
@click.option('--extensions', '-e',
              type=click.STRING,
              help="File extensions that can be stored in the CDN, example: 'jpg, png, pdf'")
@click.option('--cache-query', '-c',
              type=click.STRING,
              help="Cache query rules with the following formats:\n"
                   "'ignore-all', 'include: <query-names>', 'ignore: <query-names>'",
              default="include-all",
              show_default=True)
@environment.pass_env
def cli(env, unique_id, origin, path, origin_type, header,
        bucket_name, port, protocol, optimize_for, extensions, cache_query):
    """Create an origin path for an existing CDN mapping.

    For more information see the following documentation: \n
    https://cloud.ibm.com/docs/infrastructure/CDN?topic=CDN-manage-your-cdn#adding-origin-path-details
    """

    manager = SoftLayer.CDNManager(env.client)

    if origin_type == 'storage' and not bucket_name:
        raise exceptions.ArgumentError('[-b | --bucket-name] is required when [-t | --origin-type] is "storage"')

    result = manager.add_origin(unique_id, origin, path, origin_type=origin_type,
                                header=header, port=port, protocol=protocol,
                                bucket_name=bucket_name, file_extensions=extensions,
                                optimize_for=optimize_for, cache_query=cache_query)

    table = formatting.Table(['Item', 'Value'])
    table.align['Item'] = 'r'
    table.align['Value'] = 'r'

    table.add_row(['CDN Unique ID', result['mappingUniqueId']])

    if origin_type == 'storage':
        table.add_row(['Bucket Name', result['bucketName']])

    table.add_row(['Origin', result['origin']])
    table.add_row(['Origin Type', result['originType']])
    table.add_row(['Path', result['path']])
    table.add_row(['Port', result['httpPort']])
    table.add_row(['Configuration', result['performanceConfiguration']])
    table.add_row(['Status', result['status']])

    env.fout(table)



"""Displays information about the accounts bandwidth pools"""
# :license: MIT, see LICENSE for more details.
import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.account import AccountManager as AccountManager
from SoftLayer import utils


@click.command(cls=SLCommand, )
@environment.pass_env
def cli(env):
    """Displays bandwidth pool information

    Similiar to https://cloud.ibm.com/classic/network/bandwidth/vdr
    """

    manager = AccountManager(env.client)
    items = manager.get_bandwidth_pools()

    table = formatting.Table([
        "Id",
        "Pool Name",
        "Region",
        "Servers",
        "Allocation",
        "Current Usage",
        "Projected Usage"
    ], title="Bandwidth Pools")
    table.align = 'l'
    for item in items:
        id_bandwidth = item.get('id')
        name = item.get('name')
        region = utils.lookup(item, 'locationGroup', 'name')
        servers = manager.get_bandwidth_pool_counts(identifier=item.get('id'))
        allocation = "{} GB".format(item.get('totalBandwidthAllocated', 0))
        current = "{} GB".format(utils.lookup(item, 'billingCyclePublicBandwidthUsage', 'amountOut'))
        projected = "{} GB".format(item.get('projectedPublicBandwidthUsage', 0))

        table.add_row([id_bandwidth, name, region, servers, allocation, current, projected])
    env.fout(table)



"""
    SoftLayer.CLI.environment
    ~~~~~~~~~~~~~~~~~~~~~~~~~
    Abstracts everything related to the user's environment when running the CLI

    :license: MIT, see LICENSE for more details.
"""
import configparser
import os

import importlib
from json.decoder import JSONDecodeError

import click
import pkg_resources
from rich.console import Console
from rich.syntax import Syntax

import SoftLayer
from SoftLayer.CLI import formatting
from SoftLayer.CLI import routes
from SoftLayer import utils

# pylint: disable=too-many-instance-attributes, invalid-name

# Calling pkg_resources.iter_entry_points shows a false-positive
# pylint: disable=no-member


class Environment(object):
    """Provides access to the current CLI environment."""

    def __init__(self):
        # {'path:to:command': ModuleLoader()}
        # {'vs:list': ModuleLoader()}
        self.commands = {}
        self.aliases = {}

        self.vars = {}

        self.client = None
        self.theme = self.set_env_theme()
        self.console = utils.console_color_themes(self.theme)
        self.err_console = Console(stderr=True)
        self.format = 'table'
        self.skip_confirmations = False
        self.config_file = None

        self._modules_loaded = False

    def out(self, output):
        """Outputs a string to the console (stdout)."""

        # If we output to a | or file, need to set default width so all output is printed.
        if not self.console.is_terminal:
            self.console.width = 1000000
        if self.format == 'json':
            try:
                self.console.print_json(output)
            # Tried to print not-json, so just print it out normally...
            except JSONDecodeError:
                click.echo(output)
        elif self.format == 'jsonraw':
            #  Using Rich here is problematic because in the unit tests it thinks the terminal is 80 characters wide
            #  and only prints out that many characters.
            click.echo(output)
        else:
            # If we want to print a list of tables, Rich doens't handle that well.
            if isinstance(output, list):
                for line in output:
                    self.console.print(line, overflow='ignore')
            else:
                self.console.print(output,  overflow='ignore')

    def err(self, output, newline=True):
        """Outputs an error string to the console (stderr)."""

        self.err_console.print(output, new_line_start=newline)

    def fmt(self, output, fmt=None):
        """Format output based on current the environment format."""
        if fmt is None:
            fmt = self.format
        return formatting.format_output(output, fmt, self.theme)

    def format_output_is_json(self):
        """Return True if format output is json or jsonraw"""
        return 'json' in self.format

    def fout(self, output):
        """Format the input and output to the console (stdout)."""
        if output is not None:
            try:
                self.out(self.fmt(output))
            except UnicodeEncodeError:
                # If we hit an undecodeable entry, just try outputting as json.
                self.out(self.fmt(output, 'json'))

    def python_output(self, output):
        """Prints out python code"""
        self.console.print(Syntax(output, "python"))

    def input(self, prompt, default=None, show_default=True):
        """Provide a command prompt."""
        return click.prompt(prompt, default=default, show_default=show_default)

    def getpass(self, prompt, default=None):
        """Provide a password prompt."""
        password = click.prompt(prompt, hide_input=True, default=default)

        # https://github.com/softlayer/softlayer-python/issues/1436
        # click.prompt uses python's getpass() in the background
        # https://github.com/python/cpython/blob/3.9/Lib/getpass.py#L97
        # In windows, shift+insert actually inputs the below 2 characters
        # If we detect those 2 characters, need to manually read from the clipbaord instead
        # https://stackoverflow.com/questions/101128/how-do-i-read-text-from-the-clipboard
        if password == 'àR':
            # tkinter is a built in python gui, but it has clipboard reading functions.
            # pylint: disable=import-outside-toplevel
            from tkinter import Tk
            tk_manager = Tk()
            password = tk_manager.clipboard_get()
            # keep the window from showing
            tk_manager.withdraw()
        return password

    # Command loading methods
    def list_commands(self, *path):
        """Command listing."""
        path_str = ':'.join(path)

        commands = []
        for command in self.commands:

            # Filter based on prefix and the segment length
            if all([command.startswith(path_str),
                    len(path) == command.count(":")]):

                # offset is used to exclude the path that the caller requested.
                offset = len(path_str) + 1 if path_str else 0
                commands.append(command[offset:])

        return sorted(commands)

    def get_command(self, *path):
        """Return command at the given path or raise error."""
        path_str = ':'.join(path)

        if path_str in self.commands:
            return self.commands[path_str].load()

        return None

    def resolve_alias(self, path_str):
        """Returns the actual command name. Uses the alias mapping."""
        if path_str in self.aliases:
            return self.aliases[path_str]
        return path_str

    def load(self):
        """Loads all modules."""
        if self._modules_loaded is True:
            return

        self.load_modules_from_python(routes.ALL_ROUTES)
        self.aliases.update(routes.ALL_ALIASES)
        self._load_modules_from_entry_points('softlayer.cli')

        self._modules_loaded = True

    def load_modules_from_python(self, route_list):
        """Load modules from the native python source."""
        for name, modpath in route_list:
            if ':' in modpath:
                path, attr = modpath.split(':', 1)
            else:
                path, attr = modpath, None
            self.commands[name] = ModuleLoader(path, attr=attr)

    def _load_modules_from_entry_points(self, entry_point_group):
        """Load modules from the entry_points (slower).

        Entry points can be used to add new commands to the CLI.

        Usage:

            entry_points={'softlayer.cli': ['new-cmd = mymodule.new_cmd.cli']}

        """
        for obj in pkg_resources.iter_entry_points(group=entry_point_group,
                                                   name=None):
            self.commands[obj.name] = obj

    def ensure_client(self, config_file=None, is_demo=False, proxy=None):
        """Create a new SLAPI client to the environment.

        This will be a no-op if there is already a client in this environment.
        """
        if self.client is not None:
            return

        # Environment can be passed in explicitly. This is used for testing
        if is_demo:
            client = SoftLayer.BaseClient(
                transport=SoftLayer.FixtureTransport(),
                auth=None,
            )
        else:
            # Create SL Client
            client = SoftLayer.create_client_from_env(
                proxy=proxy,
                config_file=config_file,
            )
        self.client = client

    def set_env_theme(self, config_file=None):
        """Get theme to color console and set in env"""
        theme = os.environ.get('SL_THEME')
        if theme:
            return theme
        else:
            config_files = ['/etc/softlayer.conf', '~/.softlayer']
            path_os = os.getenv('HOME')
            if path_os:
                config_files.append(path_os + '\\AppData\\Roaming\\softlayer')
            if config_file:
                config_files.append(config_file)
            config = configparser.RawConfigParser({'theme': 'dark'})
            config.read(config_files)
            if config.has_section('softlayer'):
                self.theme = config.get('softlayer', 'theme')
                return config.get('softlayer', 'theme')
        return 'dark'


class ModuleLoader(object):
    """Module loader that acts a little like an EntryPoint object."""

    def __init__(self, import_path, attr=None):
        self.import_path = import_path
        self.attr = attr

    def load(self):
        """load and return the module/attribute."""
        module = importlib.import_module(self.import_path)
        if self.attr:
            return getattr(module, self.attr)
        return module


pass_env = click.make_pass_decorator(Environment, ensure=True)



"""Compute images."""
# :license: MIT, see LICENSE for more details.
from SoftLayer.CLI import formatting


MASK = ('id,accountId,name,globalIdentifier,parentId,publicFlag,flexImageFlag,'
        'imageType')
DETAIL_MASK = MASK + (',firstChild,children[id,blockDevicesDiskSpaceTotal,datacenter,'
                      'transaction[transactionGroup,transactionStatus]],'
                      'note,createDate,status,transaction')
PUBLIC_TYPE = formatting.FormattedItem('PUBLIC', 'Public')
PRIVATE_TYPE = formatting.FormattedItem('PRIVATE', 'Private')



"""Order a file storage replica volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers
from SoftLayer import utils


CONTEXT_SETTINGS = {'token_normalize_func': lambda x: x.upper()}


@click.command(cls=SoftLayer.CLI.command.SLCommand, context_settings=CONTEXT_SETTINGS)
@click.argument('volume_id')
@click.option('--snapshot-schedule', '-s',
              help='Snapshot schedule to use for replication, '
              '(HOURLY | DAILY | WEEKLY)',
              required=True,
              type=click.Choice(['HOURLY', 'DAILY', 'WEEKLY']))
@click.option('--location', '-l',
              help='Short name of the data center for the replicant '
              '(e.g.: dal09)',
              required=True)
@click.option('--tier',
              help='Endurance Storage Tier (IOPS per GB) of the primary'
              ' volume for which a replicant is ordered [optional]',
              type=click.Choice(['0.25', '2', '4', '10']))
@environment.pass_env
def cli(env, volume_id, snapshot_schedule, location, tier):
    """Order a file storage replica volume."""
    file_manager = SoftLayer.FileStorageManager(env.client)
    file_volume_id = helpers.resolve_id(file_manager.resolve_ids, volume_id, 'File Storage')

    if tier is not None:
        tier = float(tier)

    try:
        order = file_manager.order_replicant_volume(
            file_volume_id,
            snapshot_schedule=snapshot_schedule,
            location=location,
            tier=tier,
        )
    except ValueError as ex:
        raise exceptions.ArgumentError(str(ex))

    if 'placedOrder' in order.keys():
        click.echo("Order #{0} placed successfully!".format(
            utils.lookup(order, 'placedOrder', 'id')))
        for item in utils.lookup(order, 'placedOrder', 'items'):
            click.echo(" > %s" % item.get('description'))
    else:
        click.echo("Order could not be placed! Please verify your options " +
                   "and try again.")



"""List CDN Accounts."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--sortby',
              help='Column to sort by',
              type=click.Choice(['unique_id',
                                 'domain',
                                 'origin',
                                 'vendor',
                                 'cname',
                                 'status']))
@environment.pass_env
def cli(env, sortby):
    """List all CDN accounts."""

    manager = SoftLayer.CDNManager(env.client)
    accounts = manager.list_cdn()

    table = formatting.Table(['unique_id',
                              'domain',
                              'origin',
                              'vendor',
                              'cname',
                              'status'])
    for account in accounts:
        table.add_row([
            account['uniqueId'],
            account['domain'],
            account['originHost'],
            account['vendorName'],
            account['cname'],
            account['status']
        ])

    table.sortby = sortby
    env.fout(table)
